import numpy as np
from numpy.linalg import norm
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal

from flatspin import IsingSpinIce, CustomSpinIce

def test_init(opencl):
    si = IsingSpinIce(size=(5,4), **opencl)
    N = 5 * 4

    assert si.size == (5, 4)
    assert si.num_neighbors == 4
    assert si.spin_count == N

    assert si.spin.shape == (N,)
    assert si.pos.shape == (N, 2)
    assert si.angle.shape == (N,)
    assert si.threshold.shape == (N,)
    assert si.m.shape == (N, 2)

    assert_array_equal(si.angle, np.pi/2) # spin dir is up

    expect = np.zeros(si.m.shape)
    expect[...,1] = 1
    assert_allclose(si.m, expect)

def test_label():
    si = IsingSpinIce(size=(3,2))

    assert si.label(0) == (0,0)
    assert si.label(1) == (0,1)
    assert si.label(2) == (0,2)
    assert si.label(3) == (1,0)
    assert si.label(4) == (1,1)
    assert si.label(5) == (1,2)

def test_labels():
    si = IsingSpinIce(size=(2,2))
    assert_array_equal(si.labels, [(0,0), (0,1), (1,0), (1,1)])

    si = IsingSpinIce(size=(3,2))
    assert_array_equal(si.labels, [(0,0), (0,1), (0,2), (1,0), (1,1), (1,2)])
    assert_array_equal(si.labels[[0,2,4]], [(0,0), (0,2), (1,1)])
    assert_array_equal(si.labels[1::2], [(0,1), (1,0), (1,2)])

def test_indexof():
    si = IsingSpinIce(size=(3,2))

    inds = [si.indexof(tuple(l)) for l in si.labels]
    assert_array_equal(inds, si.indices())

    # row 1
    assert_array_equal(si.indexof(1), [3, 4, 5])

    # col 1
    assert_array_equal(si.indexof((slice(None), 1)), [1, 4])

    # every other column
    assert_array_equal(si.indexof((slice(None), slice(0, None, 2))), [0, 2, 3, 5])

def test_randomize():
    seed = 0x1234321
    si = IsingSpinIce(size=(10,10), random_seed=seed)

    assert np.all(si.spin == 1)

    si.randomize()

    assert not np.all(si.spin == 1)
    assert not np.all(si.spin == -1)
    assert np.all(np.isin(si.spin, (-1,1)))

    spin0 = si.spin.copy()
    si.randomize()
    assert not np.all(si.spin == spin0)

    si.randomize(seed=seed)
    assert np.all(si.spin == spin0)

def test_total_fields_single(opencl):
    si = IsingSpinIce(size=(1,1), **opencl)
    si.randomize()

    h_tot = si.total_fields()
    for i in si.indices():
        assert tuple(h_tot[i]) == tuple(si.total_field(i))

def test_dipolar_fields_single(opencl):
    si = IsingSpinIce(size=(1,1), **opencl)
    si.randomize()

    h_dip = si.dipolar_fields()
    for i in si.indices():
        assert tuple(h_dip[i]) == tuple(si.dipolar_field(i))

def test_L():
    si = IsingSpinIce(size=(4,3))

    inds = list(si.L[tuple(i)] for i in si.labels)

    assert_array_equal(inds, si.indices())

    assert si.L[0,0] == 0
    assert si.L[0,1] == 1
    assert si.L[0,2] == 2
    assert si.L[0,3] == 3
    assert si.L[1,0] == 4
    assert si.L[1,1] == 5
    assert si.L[1,2] == 6
    assert si.L[1,3] == 7
    assert si.L[2,0] == 8
    assert si.L[2,1] == 9
    assert si.L[2,2] == 10
    assert si.L[2,3] == 11

    # row 1
    assert_array_equal(si.L[1], [4, 5, 6, 7])

    # col 1
    assert_array_equal(si.L[:,1], [1, 5, 9])

    # rows 1:2
    assert_array_equal(si.L[1:3], [4, 5, 6, 7, 8, 9, 10, 11])

    # cols 2:3
    assert_array_equal(si.L[:,2:4], [2, 3, 6, 7, 10, 11])

    # every other row
    assert_array_equal(si.L[::2], [0, 1, 2, 3, 8, 9, 10, 11])

    # every other column
    assert_array_equal(si.L[:,::2], [0, 2, 4, 6, 8, 10])

    # rows 1:2, cols 2:3
    assert_array_equal(si.L[1:3,2:4], [6, 7, 10, 11])

    # rows 0 and 2
    assert_array_equal(si.L[[0,2]], [0, 1, 2, 3, 8, 9, 10, 11])

    # columns 1 and 3
    assert_array_equal(si.L[:,[1,3]], [1, 3, 5, 7, 9, 11])

    # last row
    assert_array_equal(si.L[-1], [8, 9, 10, 11])

    # last column
    assert_array_equal(si.L[:,-1], [3, 7, 11])

def test_pos():
    si = IsingSpinIce(size=(2,3))
    expect = [[0,0], [1,0], [0,1], [1,1], [0,2], [1,2]]
    assert_array_equal(si.pos, expect)

def test_lattice_spacing():
    L = np.pi
    si = IsingSpinIce(size=(2,3), lattice_spacing=L)
    expect = L * np.array([[0,0], [1,0], [0,1], [1,1], [0,2], [1,2]])
    assert_array_equal(si.pos, expect)

def test_threshold():
    # No disorder
    hc = 0.0142
    si = IsingSpinIce(size=(101,102), hc=hc)

    assert si.hc == hc
    assert_array_equal(si.threshold, hc)

    # With disorder
    hc = 0.1
    disorder = 0.3
    si = IsingSpinIce(size=(70,80), hc=hc, disorder=disorder)

    assert si.hc == hc
    assert np.mean(si.threshold) == approx(hc, abs=1e-2)
    assert np.std(si.threshold) == approx(disorder * hc, abs=1e-3)

    # make sure we don't have negative thresholds
    si = IsingSpinIce(size=(100,100), hc=1, disorder=100)

    assert np.count_nonzero(si.threshold < 0) == 0

def test_threshold_grid():
    # No disorder
    hc = np.array(
        [[0.1, 0.2, 0.3],
         [0.4, 0.5, 0.6]]
    )
    si = IsingSpinIce(size=(63,92), hc=hc)

    assert_array_equal(si.hc, hc)

    grid = si.fixed_grid((hc.shape[1], hc.shape[0]))
    for p in np.ndindex(grid.size):
        inds = grid.point_index(p)
        assert_array_equal(si.threshold[inds], hc[p])

    # With disorder
    disorder = 0.04
    si = IsingSpinIce(size=(103,97), hc=hc, disorder=disorder)

    assert_array_equal(si.hc, hc)

    grid = si.fixed_grid((hc.shape[1], hc.shape[0]))
    for p in np.ndindex(grid.size):
        inds = grid.point_index(p)
        assert np.mean(si.threshold[inds]) == approx(hc[p], abs=1e-2)
        assert np.std(si.threshold[inds]) == approx(disorder * hc[p], abs=1e-3)

    # make sure we don't have negative thresholds
    si = IsingSpinIce(size=(91,123), hc=hc, disorder=100)
    assert np.count_nonzero(si.threshold < 0) == 0

def test_set_hc():
    # No disorder
    hc = 0.0123
    si = IsingSpinIce(size=(100,100), hc=hc)

    assert si.hc == hc
    assert_array_equal(si.threshold, hc)

    hc = 0.0333
    si.set_hc(hc)
    assert si.hc == hc
    assert_array_equal(si.threshold, hc)

    # With disorder
    hc = 0.1
    disorder = 0.3
    si = IsingSpinIce(size=(100,100), hc=hc, disorder=disorder)

    assert si.hc == hc
    assert np.mean(si.threshold) == approx(hc, abs=1e-2)
    assert np.std(si.threshold) == approx(disorder * hc, abs=1e-3)

    old_hc = si.hc
    old_threshold = si.threshold.copy()
    hc = 0.2
    si.set_hc(hc)
    assert si.hc == hc
    assert np.mean(si.threshold) == approx(hc, abs=1e-2)
    assert np.std(si.threshold) == approx(disorder * hc, abs=1e-3)

    # Test that the new thresholds retain the distribution of the old thresholds
    assert si.threshold == approx(old_threshold * hc / old_hc)

    # no grid -> no grid
    #   no need to invoke grid
    # no grid -> grid
    #   call set_hc_grid()
    # grid -> grid
    #   call set_hc_grid()
    # grid -> no grid
    #   call set_hc_grid()


def test_set_hc_grid():
    # No disorder
    hc = np.array(
        [[0.1, 0.2, 0.3],
         [0.4, 0.5, 0.6]]
    )
    si = IsingSpinIce(size=(32, 23))

    # no grid -> grid
    si.set_hc(hc)

    grid = si.fixed_grid((hc.shape[1], hc.shape[0]))
    for p in np.ndindex(grid.size):
        inds = grid.point_index(p)
        assert_array_equal(si.threshold[inds], hc[p])

    # grid -> grid
    hc = np.array(
        [[0.3, 0.1, 0.02]]
    )
    si.set_hc(hc)

    grid = si.fixed_grid((hc.shape[1], hc.shape[0]))
    for p in np.ndindex(grid.size):
        inds = grid.point_index(p)
        assert_array_equal(si.threshold[inds], hc[p])

    # grid -> no grid
    hc = 0.132
    si.set_hc(hc)
    assert_array_equal(si.threshold, hc)


    # With disorder
    hc = np.array(
        [[0.1, 0.2, 0.3],
         [0.4, 0.5, 0.6]]
    )
    disorder = 0.05
    si = IsingSpinIce(size=(132, 123), disorder=disorder)

    # no grid -> grid
    old_hc = si.hc
    old_threshold = si.threshold.copy()
    si.set_hc(hc)

    grid = si.fixed_grid((hc.shape[1], hc.shape[0]))
    for p in np.ndindex(grid.size):
        inds = grid.point_index(p)
        assert np.mean(si.threshold[inds]) == approx(hc[p], abs=1e-2)
        assert np.std(si.threshold[inds]) == approx(disorder * hc[p], abs=1e-3)
        assert si.threshold[inds] == approx(old_threshold[inds] * hc[p] / old_hc)

    # grid -> grid
    hc = np.array(
        [[0.3, 0.1, 0.02],
         [0.6, 0.5, 0.4]]
    )
    si.set_hc(hc)

    grid = si.fixed_grid((hc.shape[1], hc.shape[0]))
    for p in np.ndindex(grid.size):
        inds = grid.point_index(p)
        assert np.mean(si.threshold[inds]) == approx(hc[p], abs=1e-2)
        assert np.std(si.threshold[inds]) == approx(disorder * hc[p], abs=1e-3)
        assert si.threshold[inds] == approx(old_threshold[inds] * hc[p] / old_hc)

    # grid -> no grid
    hc = 0.132
    si.set_hc(hc)
    assert np.mean(si.threshold) == approx(hc, abs=1e-2)
    assert np.std(si.threshold) == approx(disorder * hc, abs=1e-3)
    assert si.threshold == approx(old_threshold * hc / old_hc)


def test_num_neighbors():
    # global neighborhood
    si = IsingSpinIce(size=(3,3), neighbor_distance=np.inf)
    assert si.num_neighbors == si.spin_count - 1

    # nearest neighborhood (default)
    si = IsingSpinIce(size=(3,3))
    assert si.num_neighbors == 4

    # nearest including diagonals
    si = IsingSpinIce(size=(3,3), neighbor_distance=np.sqrt(2))
    assert si.num_neighbors == 8

def test_neighbors():
    # global neighborhood
    si = IsingSpinIce(size=(3,3), neighbor_distance=np.inf)
    indices = set(si.indices())
    for i in indices:
        assert set(si.neighbors(i)) == (indices - set([i]))

    # nearest neighborhood (default)
    si = IsingSpinIce(size=(3,3))
    L = si.L
    assert set(si.neighbors(L[0,0])) == set([L[0,1], L[1,0]])
    assert set(si.neighbors(L[0,1])) == set([L[0,0], L[0,2], L[1,1]])
    assert set(si.neighbors(L[0,2])) == set([L[0,1], L[1,2]])
    assert set(si.neighbors(L[1,0])) == set([L[0,0], L[1,1], L[2,0]])
    assert set(si.neighbors(L[1,1])) == set([L[0,1], L[1,0], L[1,2], L[2,1]])
    assert set(si.neighbors(L[1,2])) == set([L[0,2], L[1,1], L[2,2]])
    assert set(si.neighbors(L[2,0])) == set([L[1,0], L[2,1]])
    assert set(si.neighbors(L[2,1])) == set([L[1,1], L[2,0], L[2,2]])
    assert set(si.neighbors(L[2,2])) == set([L[1,2], L[2,1]])


def test_spin_dipolar_field(opencl):
    # spins point up
    si = IsingSpinIce(size=(2,2), **opencl)
    L = si.L

    j_T = 1
    j_L = 2
    j_D = 1/(4*np.sqrt(2))

    # parallell spins
    assert si.spin_dipolar_field(L[0,0], L[0,1]) == approx([-j_T, 0])

    # aligned spins
    assert si.spin_dipolar_field(L[0,0], L[1,0]) == approx([j_L, 0])

    # 45 degree neighbors
    assert si.spin_dipolar_field(L[0,0], L[1,1]) == approx([j_D, -3*j_D])

def test_dipolar_field(opencl):
    si = IsingSpinIce(size=(2,2), alpha=1.0, neighbor_distance=np.inf, **opencl)
    L = si.L

    j_T = 1
    j_L = 2
    j_D = 1/(4*np.sqrt(2))

    assert si.dipolar_field(L[0,0]) == approx([-j_T + j_L + j_D, -3*j_D])
    si.flip(L[1,0])
    assert si.dipolar_field(L[0,0]) == approx([-j_T - j_L + j_D, -3*j_D])
    si.flip(L[0,0])
    assert si.dipolar_field(L[0,0]) == approx([j_T + j_L - j_D, 3*j_D])

    si = IsingSpinIce(size=(4,4), alpha=1.0, **opencl)

    for i in si.indices():
        tot = np.sum([si.spin_dipolar_field(i, j) for j in si.neighbors(i)], axis=0)
        assert tuple(si.dipolar_field(i)) == approx(tot)

def test_external_field(opencl):
    si = IsingSpinIce(size=(2,2), **opencl)
    L = si.L

    h_par, h_perp = si.external_field(L[0,0])
    assert h_par == h_perp == 0

    # global external field
    si.set_h_ext((0,1))
    h_par, h_perp = si.external_field(L[0,0])
    assert h_par == 1
    assert h_perp == approx(0)

    si.flip(L[0,0])
    assert si.external_field(L[0,0]) == approx([-1,0])

    si.set_h_ext((1,0))
    assert si.external_field(L[0,0]) == approx([0,1])

    # local external field
    h_ext = [(0,1), (1,0),
             (0,2), (-2,0)]
    si.set_h_ext(h_ext)
    si.polarize()

    assert si.external_field(L[0,0]) == approx([1, 0])
    assert si.external_field(L[0,1]) == approx([0, -1])
    assert si.external_field(L[1,0]) == approx([2, 0])
    assert si.external_field(L[1,1]) == approx([0, 2])

def test_total_field(opencl):
    si = IsingSpinIce(size=(4,4), **opencl)
    si.randomize()

    # global external field
    si.set_h_ext((0,1))

    i = si.L[2,2]
    assert tuple(si.total_field(i)) == tuple(si.dipolar_field(i) + si.external_field(i))

    # local external field
    h_ext = np.zeros((si.spin_count, 2))
    h_ext[i] = [0,2]
    si.set_h_ext(h_ext)
    assert tuple(si.total_field(i)) == tuple(si.dipolar_field(i) + si.external_field(i))

def test_dipolar_fields(opencl):
    si = IsingSpinIce(size=(5,7), **opencl)
    si.randomize()

    h_dip = si.dipolar_fields()
    for i in si.indices():
        assert tuple(h_dip[i]) == tuple(si.dipolar_field(i))

def test_external_fields(opencl):
    si = IsingSpinIce(size=(3,9), **opencl)
    si.randomize()

    # global external field
    si.set_h_ext((2,5))

    h_ext = si.external_fields()
    for i in si.indices():
        assert tuple(h_ext[i]) == approx(tuple(si.external_field(i)))

    # local external field
    h_ext = np.random.uniform(-10, 10, (si.spin_count, 2))
    si.set_h_ext(h_ext)

    h_ext = si.external_fields()
    for i in si.indices():
        assert tuple(h_ext[i]) == approx(tuple(si.external_field(i)))

def test_total_fields(opencl):
    si = IsingSpinIce(size=(1,15), **opencl)
    si.randomize()

    # global external field
    si.set_h_ext((0,1))

    h_tot = si.total_fields()
    h_dip = si.dipolar_fields()
    h_ext = si.external_fields()
    assert_array_equal(h_tot, h_ext + h_dip)

    # local external field
    h_ext = np.random.uniform(-10, 10, (si.spin_count, 2))
    si.set_h_ext(h_ext)

    h_tot = si.total_fields()
    h_dip = si.dipolar_fields()
    h_ext = si.external_fields()
    assert_allclose(h_tot, h_ext + h_dip, 6)

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_flippable(opencl, switching):
    si = IsingSpinIce(size=(4,4), alpha=1.0, hc=1.0, sw_b=1, neighbor_distance=np.inf, switching=switching, **opencl)
    L = si.L
    assert len(si.flippable()) == 0
    si.flip(L[0,0])
    expected = [L[0,0]]
    assert_array_equal(si.flippable(), expected)

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_alpha(opencl, switching):
    hc = 1.0
    h_ext = -(hc + 2)
    si = IsingSpinIce(size=(4,4), alpha=1.0, hc=hc, h_ext=(0, h_ext), switching=switching, **opencl)
    h_dip = si.dipolar_fields()
    h_external = si.external_fields()
    h_tot = si.total_fields()
    n_flip = len(si.flippable())

    assert 0 < n_flip < si.spin_count

    # adjust alpha only
    alpha = 0.5
    si.set_alpha(alpha)

    assert_array_equal(si.dipolar_fields(), alpha * h_dip)
    assert_array_equal(si.external_fields(), h_external)
    assert_array_equal(si.total_fields(), alpha * h_dip + h_external)

    # adjust alpha, rescale hc and h_ext
    alpha = 0.5
    hc *= alpha
    h_ext *= alpha
    si = IsingSpinIce(size=(4,4), hc=hc, h_ext=(0, h_ext), alpha=alpha, switching=switching, **opencl)

    assert_array_equal(si.dipolar_fields(), alpha * h_dip)
    assert_array_equal(si.total_fields(), alpha * h_tot)
    assert len(si.flippable()) == n_flip

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_step(opencl, switching):
    si = IsingSpinIce(size=(4,4), alpha=1.0, hc=1.0, sw_b=1, neighbor_distance=np.inf, switching=switching, **opencl)
    si.flip(0)
    assert len(si.flippable()) == 1
    assert si.step()
    assert len(si.flippable()) == 0

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_relax(opencl, switching):
    si = IsingSpinIce(size=(8,8), alpha=1.0, hc=1.0, neighbor_distance=np.inf, switching=switching, **opencl)
    L = si.L
    assert si.relax() == 0
    si.flip(L[0,0])
    si.flip(L[0,1])
    assert si.relax() == 2
    assert len(si.flippable()) == 0

def test_relax_single(opencl):
    #if "use_cuda" in opencl:
    #    pytest.skip("test breaks subsequent cuda calls (memory corruption?)")

    si = IsingSpinIce(size=(1,1), alpha=1.0, hc=1.0, neighbor_distance=np.inf, **opencl)
    L = si.L
    assert si.relax() == 0
    si.set_h_ext((0,-1))
    assert si.relax() == 1
    assert len(si.flippable()) == 0

def test_energy(opencl):
    si = IsingSpinIce(size=(4,4), hc=1.0, **opencl)
    E0 = si.energy()
    si.flip(0)
    E1 = si.energy()
    assert E1[0] > E0[0]

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_total_energy(opencl, switching):
    si = IsingSpinIce(size=(6,5), hc=1.0, neighbor_distance=np.inf, switching=switching, **opencl)

    E = si.total_energy()
    si.flip(0)
    assert si.total_energy() > E

    si.randomize()
    E = si.total_energy()
    while si.step():
        assert si.total_energy() < E
        E = si.total_energy()

def test_vectors():
    si = IsingSpinIce(size=(4,2))
    spin = [1,  1, 1, -1,
            1, -1, -1, 1]
    si.set_spin(spin)

    # grid_spacing == lattice_spacing
    vectors = si.vectors
    expect = [[0,1], [0,1],  [0, 1], [0,-1],
              [0,1], [0,-1], [0,-1], [0, 1]]
    assert_allclose(vectors, expect)

def test_find_vertices():
    si = IsingSpinIce(size=(2,3))
    L = si.L
    vi, vj, indices = si.find_vertices()

    assert_array_equal(vi, [0, 0, 1, 1])
    assert_array_equal(vj, [0, 1, 0, 1])

    assert tuple(indices[0]) == (L[0,0], L[1,0])
    assert tuple(indices[1]) == (L[0,1], L[1,1])
    assert tuple(indices[2]) == (L[1,0], L[2,0])
    assert tuple(indices[3]) == (L[1,1], L[2,1])

def test_vertices():
    si = IsingSpinIce(size=(2,3))
    L = si.L
    indices = si.vertices()
    assert_array_equal(indices,
        [[L[0,0], L[1,0]],
         [L[0,1], L[1,1]],
         [L[1,0], L[2,0]],
         [L[1,1], L[2,1]]])

def test_vertex_indices():
    si = IsingSpinIce(size=(2,3))
    indices = np.transpose(si.vertex_indices())
    assert_array_equal(indices, [(0,0), (0,1), (1,0), (1,1)])

def test_vertex_type():
    si = IsingSpinIce(size=(4,2))
    spin = [1, -1, 1, -1,
            1, -1, -1, 1]
    si.set_spin(spin)

    vertex = si.vertices()

    assert si.vertex_type(vertex[0]) == 1
    assert si.vertex_type(vertex[1]) == 1
    assert si.vertex_type(vertex[2]) == 2
    assert si.vertex_type(vertex[3]) == 2

def test_vertex_pos():
    si = IsingSpinIce(size=(3,3))

    vertex = si.vertices()

    assert tuple(si.vertex_pos(vertex[0])) == (0, 0.5)
    assert tuple(si.vertex_pos(vertex[1])) == (1, 0.5)
    assert tuple(si.vertex_pos(vertex[2])) == (2, 0.5)
    assert tuple(si.vertex_pos(vertex[3])) == (0, 1.5)
    assert tuple(si.vertex_pos(vertex[4])) == (1, 1.5)
    assert tuple(si.vertex_pos(vertex[5])) == (2, 1.5)

def test_vertex_count():
    si = IsingSpinIce(size=(4,2))
    spin = [1, -1, -1, -1,
            1, -1, -1, 1]
    si.set_spin(spin)

    types, counts = si.vertex_count()
    assert types == (1, 2)
    assert counts == (3, 1)

def test_vertex_population():
    si = IsingSpinIce(size=(4,2))
    spin = [1, -1, 1, -1,
            1, -1, -1, 1]
    si.set_spin(spin)

    types, pops = si.vertex_population()
    assert types == (1, 2)
    assert pops == (0.5, 0.5)

def test_vertex_mag():
    si = IsingSpinIce(size=(4,2))
    spin = [1, -1, 1, -1,
            1, -1, -1, 1]
    si.set_spin(spin)

    vertex = si.vertices()

    assert_allclose(si.vertex_mag(vertex[0]), [0, 2])
    assert_allclose(si.vertex_mag(vertex[1]), [0, -2])
    assert_allclose(si.vertex_mag(vertex[2]), [0, 0])
    assert_allclose(si.vertex_mag(vertex[3]), [0, 0])

def test_spin_grid():
    si = IsingSpinIce(size=(4,2))
    spin = [1,  1, 1, -1,
            1, -1, -1, 1]
    si.set_spin(spin)

    assert si.lattice_spacing == 1

    # grid_spacing == lattice_spacing
    grid = si.spin_grid()
    assert grid.shape == (2, 4, 2)
    expect = [[[0,1], [0,1],  [0, 1], [0,-1]],
              [[0,1], [0,-1], [0,-1], [0, 1]]]
    assert_allclose(grid, expect)

    # grid_spacing < lattice_spacing
    grid = si.spin_grid(0.5)
    assert grid.shape == (3, 7, 2)
    expect = [[[0,1], [0,0], [0, 1], [0,0], [0, 1], [0,0], [0,-1]],
              [[0,0], [0,0], [0, 0], [0,0], [0, 0], [0,0], [0, 0]],
              [[0,1], [0,0], [0,-1], [0,0], [0,-1], [0,0], [0, 1]]]
    assert_allclose(grid, expect)

    # grid_spacing > lattice_spacing
    grid = si.spin_grid(2.0)
    assert grid.shape == (1, 2, 2)
    expect = [[[0,2], [0,0]]]
    assert_allclose(grid, expect)

    # sum of all spins
    grid = si.spin_grid(4.0)
    assert grid.shape == (1, 1, 2)
    expect = [[[0,2]]]
    assert_allclose(grid, expect)

def test_grid():
    si = IsingSpinIce(size=(4,4))
    L = si.L

    assert si.lattice_spacing == 1

    #
    # default grid fits perfectly
    #
    grid = si.grid()
    assert grid.cell_size == (1, 1)
    assert grid.size == (4, 4)

    assert grid.grid_index(L[0,0]) == (0,0)
    spin_indices = si.all_indices()
    grid_indices = grid.grid_index(spin_indices)
    assert_array_equal(grid_indices, np.transpose(si.labels))


    #
    # cell_size 2x2 evenly divides lattice spacing
    #
    grid = si.grid(cell_size=2)
    assert grid.cell_size == (2, 2)
    assert grid.size == (2, 2)

    # cell (0,0)
    i = [L[0,0], L[0,1], L[1,0], L[1,1]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [0,0])
    assert_array_equal(grid.point_index((0,0)), i)

    # cell (0,1)
    i = [L[0,2], L[0,3], L[1,2], L[1,3]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [0,1])
    assert_array_equal(grid.point_index((0,1)), i)

    # cell (1,0)
    i = [L[2,0], L[2,1], L[3,0], L[3,1]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [1,0])
    assert_array_equal(grid.point_index((1,0)), i)

    # cell (1,1)
    i = [L[2,2], L[2,3], L[3,2], L[3,3]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [1,1])
    assert_array_equal(grid.point_index((1,1)), i)


    #
    # cell_size 3x2 does not evently divide lattice_spacing
    #
    grid = si.grid(cell_size=(3, 2))
    assert grid.cell_size == (3, 2)
    assert grid.size == (2, 2)

    # cell (0,0)
    i = [L[0,0], L[0,1], L[0,2], L[1,0], L[1,1], L[1,2]]
    pi = grid.point_index((0,0))
    assert_array_equal(pi, i)

    # cell (0,1)
    i = [L[0,3], L[1,3]]
    pi = grid.point_index((0,1))
    assert_array_equal(pi, i)

    # cell (1,0)
    i = [L[2,0], L[2,1], L[2,2], L[3,0], L[3,1], L[3,2]]
    pi = grid.point_index((1,0))
    assert_array_equal(pi, i)

    # cell (1,1)
    i = [L[2,3], L[3,3]]
    pi = grid.point_index((1,1))
    assert_array_equal(pi, i)

def test_fixed_grid():
    si = IsingSpinIce(size=(4,4))
    L = si.L

    assert si.lattice_spacing == 1

    #
    # 4x4 grid fits perfectly
    #
    grid = si.fixed_grid((4, 4))
    assert grid.cell_size == (1, 1)
    assert grid.size == (4, 4)

    assert grid.grid_index(L[0,0]) == (0,0)
    spin_indices = si.all_indices()
    grid_indices = grid.grid_index(spin_indices)
    assert_array_equal(grid_indices, np.transpose(si.labels))

    #
    # 2x2 grid evenly divides lattice spacing
    #
    grid = si.fixed_grid((2,2))
    assert grid.cell_size == (2, 2)
    assert grid.size == (2, 2)

    # cell (0,0)
    i = [L[0,0], L[0,1], L[1,0], L[1,1]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [0,0])
    assert_array_equal(grid.point_index((0,0)), i)

    # cell (0,1)
    i = [L[0,2], L[0,3], L[1,2], L[1,3]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [0,1])
    assert_array_equal(grid.point_index((0,1)), i)

    # cell (1,0)
    i = [L[2,0], L[2,1], L[3,0], L[3,1]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [1,0])
    assert_array_equal(grid.point_index((1,0)), i)

    # cell (1,1)
    i = [L[2,2], L[2,3], L[3,2], L[3,3]]
    gi = grid.grid_index(i)
    assert np.all(np.transpose(gi) == [1,1])
    assert_array_equal(grid.point_index((1,1)), i)


    #
    # 3x2 grid does not evently divide lattice_spacing
    #
    grid = si.fixed_grid((3,2))
    assert grid.cell_size == (4/3, 2)
    assert grid.size == (2, 3)

    # cell (0,0)
    i = [L[0,0], L[1,0]]
    pi = grid.point_index((0,0))
    assert_array_equal(pi, i)

    # cell (0,1)
    i = [L[0,1], L[0,2], L[1,1], L[1,2]]
    pi = grid.point_index((0,1))
    assert_array_equal(pi, i)

    # cell (0,2)
    i = [L[0,3], L[1,3]]
    pi = grid.point_index((0,2))
    assert_array_equal(pi, i)

    # cell (1,0)
    i = [L[2,0], L[3,0]]
    pi = grid.point_index((1,0))
    assert_array_equal(pi, i)

    # cell (1,1)
    i = [L[2,1], L[2,2], L[3,1], L[3,2]]
    pi = grid.point_index((1,1))
    assert_array_equal(pi, i)

    # cell (1,2)
    i = [L[2,3], L[3,3]]
    pi = grid.point_index((1,2))
    assert_array_equal(pi, i)

def test_data_order():
    # Test that internal arrays are always in C order. The user might pass in
    # arrays in F order, which will cause problems with opencl which very much
    # assumes C order.
    #
    # Data order only applies to arrays with ndim > 1.  Currently the only such
    # array that is allowed to be modified by the user is h_ext.

    h_ext = np.random.uniform(size=(2, 4*4))
    h_ext = h_ext.T
    assert np.isfortran(h_ext)

    si = IsingSpinIce(size=(4,4), h_ext=h_ext)
    assert not np.isfortran(si.h_ext)

    si = IsingSpinIce(size=(4,8))
    assert not np.isfortran(si.h_ext)
    h_ext = np.random.uniform(size=(2, 4*8)).T
    assert np.isfortran(h_ext)
    si.set_h_ext(h_ext)
    assert not np.isfortran(si.h_ext)

def test_set_angle(opencl):
    params = dict(
        size=(10,10),
        neighbor_distance=3
    )
    params.update(opencl)

    model = IsingSpinIce(**params)

    # Create a CustomSpinIce with expected result
    angle = np.random.uniform(0, 360, model.spin_count)
    angle = np.deg2rad(angle)
    custom = CustomSpinIce(magnet_coords=model.pos, magnet_angles=angle, radians=True, **params)

    assert_array_equal(model.pos, custom.pos)
    assert not np.all(model.angle == custom.angle)
    assert not np.all(model.m == custom.m)
    assert not np.all(model.dipolar_fields() == custom.dipolar_fields())

    # Change angles of model
    model.set_angle(angle)

    assert_array_equal(model.pos, custom.pos)
    assert_array_equal(model.angle, custom.angle)
    assert_array_equal(model.m, custom.m)
    assert_array_equal(model.dipolar_fields(), custom.dipolar_fields())

def test_set_pos(opencl):
    params = dict(
        size=(10,10),
        neighbor_distance=3
    )
    params.update(opencl)

    model = IsingSpinIce(**params)

    # Create a CustomSpinIce with expected result
    pos = model.pos + .3*np.random.uniform(-1, 1, model.pos.shape)
    custom = CustomSpinIce(magnet_coords=pos, magnet_angles=model.angle, radians=True, **params)

    assert not np.all(model.pos == custom.pos)
    assert_array_equal(model.angle, custom.angle)
    assert_array_equal(model.m, custom.m)
    assert not np.all(model.dipolar_fields() == custom.dipolar_fields())

    # Change positions of model
    model.set_pos(pos)

    assert_array_equal(model.pos, custom.pos)
    assert_array_equal(model.angle, custom.angle)
    assert_array_equal(model.m, custom.m)
    assert_array_equal(model.dipolar_fields(), custom.dipolar_fields())

def test_set_geometry(opencl):
    params = dict(
        size=(10,10),
        neighbor_distance=3
    )
    params.update(opencl)

    model = IsingSpinIce(**params)

    # Create a CustomSpinIce with expected result
    pos = model.pos + .3*np.random.uniform(-1, 1, model.pos.shape)
    angle = np.random.uniform(0, 360, model.spin_count)
    angle = np.deg2rad(angle)
    custom = CustomSpinIce(magnet_coords=pos, magnet_angles=angle, radians=True, **params)

    assert not np.all(model.pos == custom.pos)
    assert not np.all(model.angle == custom.angle)
    assert not np.all(model.m == custom.m)
    assert not np.all(model.dipolar_fields() == custom.dipolar_fields())

    # Change positions and angles of model
    model.set_geometry(pos, angle)

    assert_array_equal(model.pos, custom.pos)
    assert_array_equal(model.angle, custom.angle)
    assert_array_equal(model.m, custom.m)
    assert_array_equal(model.dipolar_fields(), custom.dipolar_fields())

#                                              NN,        NNN, 3NN,        4NN,          5NN
@pytest.mark.parametrize("neighbor_distance", [ 0.1, 1, np.sqrt(2),   2, np.sqrt(5), 2*np.sqrt(2)])
def test_set_neighbor_distance(opencl, neighbor_distance):
    # nearest neighbors (default)
    params = dict(size=(10, 10))
    params.update(opencl)
    si = IsingSpinIce(**params)

    for i in np.arange(0.1, 10):
        si.dipolar_fields()
        si.set_neighbor_distance(i)

    expect = IsingSpinIce(neighbor_distance=neighbor_distance, **params)
    si.set_neighbor_distance(neighbor_distance)
    si.total_fields()

    assert si.neighbor_distance == expect.neighbor_distance
    assert si.num_neighbors == expect.num_neighbors
    assert_array_equal(si._neighbor_list, expect._neighbor_list)
    assert_array_equal(si.dipolar_fields(), expect.dipolar_fields())

def test_zero_neighbors(opencl):
    si = IsingSpinIce(neighbor_distance=0, **opencl)
    assert si.num_neighbors == 0
    assert all([len(si.neighbors(i)) == 0 for i in si.indices()])
    assert_array_equal(si.total_fields(), np.zeros((si.spin_count, 2)))
    assert_array_equal(si.dipolar_fields(), np.zeros((si.spin_count, 2)))

    si.set_neighbor_distance(0.1)
    assert si.num_neighbors == 0
    assert all([len(si.neighbors(i)) == 0 for i in si.indices()])
    assert_array_equal(si.total_fields(), np.zeros((si.spin_count, 2)))
    assert_array_equal(si.dipolar_fields(), np.zeros((si.spin_count, 2)))

    si.set_neighbor_distance(1)
    assert si.num_neighbors > 0
    assert all([len(si.neighbors(i)) > 0 for i in si.indices()])
    h_tot = si.total_fields()
    assert all(norm(h_tot, axis=0) > 0)
    h_dip = si.dipolar_fields()
    assert all(norm(h_dip, axis=0) > 0)



"""
Vertex analysis
"""
import numpy as np
from numpy.linalg import norm
from skimage.util import view_as_windows

def find_vertices(grid, pos, angle, win_size):
    """
    Find the vertices of a geometry

    Parameters
    ----------
    grid : Grid object
        The grid of the spin positions
    angle : 1D array
        The angles of each spin
    win_size : (height, width)
        The window size to scan the grid

    Returns a tuple (vi, vj, indices) where vi, vj are the vertex indices
    and indices is a list of spin indices corresponding to each vertex
    index.
    """

    # Map indices onto grid
    index_grid = grid.map_values(np.arange(len(angle)), -1)

    # Sliding window over index grid
    index_window = view_as_windows(index_grid, win_size)

    # Map of vertex candidates in index_window
    is_vertex = np.zeros(index_window.shape[:2], dtype=bool)

    # Average distance from magnets to vertex center
    vertex_dist = np.zeros(index_window.shape[:2])

    # Find all windows where all spins point either towards or away from the
    # window center
    mag = np.column_stack((np.cos(angle), np.sin(angle)))
    for i in np.ndindex(index_window.shape[:2]):
        inds = index_window[i].flatten()
        inds = inds[inds >= 0]

        if len(inds) == 0:
            continue

        vertex_pos = np.mean(pos[inds], axis=0)
        pos_vec = pos[inds] - vertex_pos
        vec_dirs = [np.dot(p, m) for (p,m) in zip(pos_vec, mag[inds])]
        is_vertex[i] = np.all(np.abs(vec_dirs) > 1e-8)
        vertex_dist[i] = np.mean(norm(pos_vec, axis=1))


    # For some geometries, e.g., kagome, the above search will produce vertex
    # candiates with different numbers of spins
    # Assume all vertices have the same number of spins, and keep only the
    # candidates with the most spins

    # Find number of spins in each vertex candidate
    vertex_size = index_window >= 0
    vertex_size = vertex_size.reshape(vertex_size.shape[:2] + (-1,))
    vertex_size = np.count_nonzero(vertex_size, axis=-1)

    # Keep only the vertices with the most spins
    max_size = np.max(vertex_size[is_vertex])
    is_vertex = np.logical_and(is_vertex, vertex_size == max_size)

    # Keep only the most compact vertices
    vertex_dist = np.round(vertex_dist, 12)
    min_dist = np.min(vertex_dist[is_vertex])
    is_vertex = np.logical_and(is_vertex, vertex_dist == min_dist)

    # The vertex indices (vi, vj) are based on the grid
    vertex_indices = is_vertex.nonzero()

    # Extract the spin indices for each vertex
    spin_indices = index_window[is_vertex]
    spin_indices = [i[i>=0] for i in spin_indices]

    return vertex_indices + (spin_indices,)

def vertex_type_ising(spin, pos, angle):
    if spin[0] == spin[1]:
        return 1
    return 2

def vertex_dir(pos, angle):
    """
    Calculate the direction of spins in a vertex

    Determines whether spins point towards or away from the vertex center

    Returns an array containing 1 if the spin points towards, and -1 if the
    spin points away from the vertex center, respectively """

    vertex_pos = np.mean(pos, axis=0)
    pos_vec = pos - vertex_pos
    mag_vec = np.column_stack((np.cos(angle), np.sin(angle)))
    dir_in = [np.dot(p, m) < 0 for (p,m) in zip(pos_vec, mag_vec)]
    dir_in = np.where(dir_in, 1, -1)

    return dir_in


def vertex_type_square(spin, pos, angle):
    dir_in = vertex_dir(pos, angle)
    spin = spin * dir_in
    if spin[0] == -1:
        # flipping the spins in a vertex won't change its type
        spin *= -1
    num_in = np.count_nonzero(spin == 1)

    if num_in == 2:
        if tuple(spin) == (1, -1, -1, 1):
            # Type I
            return 1
        # Type II
        return 2

    if num_in == 3 or num_in == 1:
        # Type III
        return 3

    # Type IV
    return 4

def vertex_type_tri(spin, pos, angle):
    dir_in = vertex_dir(pos, angle)
    spin *= dir_in
    if spin[0] == -1:
        # flipping the spins in a vertex won't change its type
        spin *= -1

    num_in = np.count_nonzero(spin == 1)
    if num_in == 3:
        # Type II
        return 2

    # Type I
    return 1

def vertex_type(spin, pos, angle):
    """ Determine the type of a vertex given its spins and angles """
    if len(spin) == 2:
        return vertex_type_ising(spin, pos, angle)
    elif len(spin) == 3:
        return vertex_type_tri(spin, pos, angle)
    elif len(spin) == 4:
        return vertex_type_square(spin, pos, angle)

    raise NotImplementedError(f"Don't know about vertices of size {len(spin)}, sorry.")

def vertex_pos(pos, vertices):
    return np.array([pos[v].mean(axis=0) for v in vertices])

def vertex_mag(mag, vertices):
    return np.array([mag[v].sum(axis=0) for v in vertices])




"""
Grid
"""
import numpy as np
from numba import jit

@jit(nopython=True)
def add_values_fast(grid_values, grid_inds, values):
    # add values to the grid
    for i in range(len(grid_inds)):
        gi, gj = grid_inds[i]
        grid_values[gi, gj] += values[i]

class Grid:
    def __init__(self, points, cell_size=None, padding=None):
        """ Create a regular grid over points

        Allows quick lookup of point index -> grid index

        Parameters:
        -----------
        points : array
            List of points to place on the grid
        cell_size : tuple
            Size og each cell on the grid
            If not provided, attempts to auto-detect a suitable cell_size such
            that each cell contains at most one point.
        padding : tuple
            Offset each point by a padding to ensure they don't fall exactly on
            a cell edge
        """
        if cell_size is None:
            # auto-detect
            X = np.around(points[...,0], 12) # 12 seems to work...
            Y = np.around(points[...,1], 12)

            X = np.unique(X)
            Y = np.unique(Y)

            # determine smallest spacing
            dX = np.min(X[1:] - X[:-1]) if len(X) > 1 else 1
            dY = np.min(Y[1:] - Y[:-1]) if len(Y) > 1 else 1

            cell_size = (dX, dY)

        if np.isscalar(cell_size):
            cell_size = (cell_size, cell_size)

        self.cell_size = cell_size

        if padding is None:
            padding = (cell_size[0]/2, cell_size[1]/2)

        if np.isscalar(padding):
            padding = (padding, padding)

        self.padding = padding

        cell_size = np.array(cell_size)
        padding = np.array(padding)

        # In case of negative points
        pos = points - np.min(points, axis=0)

        # Lookup table of point index to grid index
        self._grid_index = ((pos + padding) / cell_size).astype(int)

        # Flip x,y -> y,x to get (row, col)
        self._grid_index = np.flip(self._grid_index, axis=-1)

        # Calculate size of grid (nrows, ncols)
        flatindex = self._grid_index.reshape((-1, self._grid_index.shape[-1]))
        self.size = tuple(np.max(flatindex, axis=0) + 1)

        # Calculate the extent (bounding box)
        xmin = np.min(points[...,0])
        ymin = np.min(points[...,1])
        width = (self.size[1] * cell_size[0])
        height = (self.size[0] * cell_size[1])
        x0 = xmin - padding[0]
        y0 = ymin - padding[1]

        self.extent = (x0, y0, x0 + width, y0 + height)

    def centers(self):
        """ Return the centers of the grid cells """
        x0, y0, width, height = self.extent
        x = np.arange(x0 + self.cell_size[0] / 2, width, self.cell_size[0])
        y = np.arange(y0 + self.cell_size[1] / 2, height, self.cell_size[1])
        return x, y

    def edges(self):
        """ Return the edges of the grid cells """
        x0, y0, width, height = self.extent
        x = np.linspace(x0, width, self.size[1]+1)
        y = np.linspace(y0, height, self.size[0]+1)
        return x, y

    def center_grid(self):
        """ Make a grid with the cell centers """
        x, y = self.centers()
        xx, yy = np.meshgrid(x, y)
        return np.stack((xx, yy), axis=-1)

    @classmethod
    def fixed_grid(cls, pos, grid_size):
        """ Make a grid with a fixed size """
        X = np.unique(np.round(pos[...,0], 12))
        Y = np.unique(np.round(pos[...,1], 12))

        # Determine padding based on smallest spacing
        dX = np.min(X[1:] - X[:-1]) if len(X) > 1 else 1
        dY = np.min(Y[1:] - Y[:-1]) if len(Y) > 1 else 1
        padding = (dX/2, dY/2)

        width = (np.max(X) - np.min(X)) + 2 * padding[0]
        height = (np.max(Y) - np.min(Y)) + 2 * padding[1]
        cell_size = (width / grid_size[0]), (height / grid_size[1])

        padx = min(padding[0], cell_size[0]/2)
        pady = min(padding[1], cell_size[1]/2)
        padding = (padx, pady)

        return cls(pos, cell_size, padding)

    def grid_index(self, point_inds):
        """ Map point index to grid index

        If point_inds is a single index i, a single grid index (gi, gj) is
        returned.

        If point_inds is a list of indices, it should take the form of an
        integer array index: ([i0, i1, i2, ...], [j0, j1, j2, ...])
        The corresponding grid indices are returned in the same format.

        See https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#integer-array-indexing
        """
        gi = self._grid_index[point_inds].T
        return tuple(gi)

    def point_index(self, grid_index):
        """ Map one grid index to zero or more point indices """
        return np.flatnonzero(np.all(self._grid_index == grid_index, axis=-1))

    def map_values(self, values, fill_value=0, mask_empty=False):
        """ Map values onto the grid

        Returns a 2D array with the values mapped onto the grid
        """
        values = np.array(values)
        indices = np.arange(len(values))
        grid_inds = self.grid_index(indices)

        # start with an empty grid
        grid_size = self.size
        if len(values.shape) > 1:
            # e.g. for vectors
            grid_size = grid_size + values.shape[1:]

        grid_values = np.full(grid_size, fill_value, dtype=values.dtype)

        # map values onto the grid
        grid_values[grid_inds] = values

        if mask_empty:
            mask = np.ones(grid_size, dtype=bool)
            mask[grid_inds] = False
            grid_values = np.ma.array(grid_values, mask=mask)

        return grid_values

    def add_values(self, values, fill_value=0, mask_empty=False, method='sum'):
        """ Add values to the grid

        Returns a 2D array with the values summed to the grid

        Parameters:
        -----------
        values : array or list
            Values to map onto the grid cells
        fill_value : scalar
            Fill value for empty cells
        method : {'sum', 'mean'}
            How to aggregate multiple values which map to the same cell
        """
        values = np.array(values)
        indices = np.arange(len(values))
        grid_inds = self.grid_index(indices)

        # start with an empty grid
        grid_size = self.size
        if len(values.shape) > 1:
            # e.g. for vectors
            grid_size = grid_size + values.shape[1:]

        grid_values = np.full(grid_size, fill_value, dtype=values.dtype)

        # add values to the grid
        add_values_fast(grid_values, np.transpose(grid_inds), values)

        if method == 'mean':
            # calculate number of values added to each cell
            unique, num_values = np.unique(np.transpose(grid_inds), return_counts=True, axis=0)
            u = tuple(unique.T)
            if len(num_values.shape) < len(grid_values[u].shape):
                num_values.shape += (1,)

            grid_values[u] = grid_values[u] / num_values

        if mask_empty:
            mask = np.ones(grid_size, dtype=bool)
            mask[grid_inds] = False
            grid_values = np.ma.array(grid_values, mask=mask)

        return grid_values



import numpy as np
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal
from itertools import groupby

from flatspin import LatticeSpinIce

def test_init(opencl):
    b0, b1 = [1,0], [0,1]
    a=0.5
    si = LatticeSpinIce(size=(3,3), lattice_spacing=a, basis0=b0, basis1=b1, const_angle=45, radians=False, **opencl)

    assert si.size == (3,3)
    assert si.spin_count == 3 * 3
    assert si.num_neighbors == 4

    assert_array_equal(si.pos, np.array([[0. , 0. ],
                                        [0.5, 0. ],
                                        [1. , 0. ],
                                        [0. , 0.5],
                                        [0.5, 0.5],
                                        [1. , 0.5],
                                        [0. , 1. ],
                                        [0.5, 1. ],
                                        [1. , 1. ]]))
    assert_array_equal(si.angle, [np.deg2rad(45)] * 9)

    si = LatticeSpinIce(size=(3,3), lattice_spacing=a, basis0=b0, basis1=b1, const_angle=np.pi, radians=True, **opencl)
    assert_array_equal(si.angle, [np.pi] * 9)

def test_labels():
    si = LatticeSpinIce(size=(2,2))
    assert_array_equal(si.labels,
            [[0, 0],[1, 0],
             [0, 1],[1, 1]])
    si = LatticeSpinIce(size=(3,4))
    assert_array_equal(si.labels,
            [[0, 0],[1, 0],[2, 0],
             [0, 1],[1, 1],[2, 1],
             [0, 2],[1, 2],[2, 2],
             [0, 3],[1, 3],[2, 3 ]])

def test_set_basis(opencl):
    b0, b1 = [1,0], [0,1]
    a = 0.5

    params = dict(
            size=(3,3), lattice_spacing=a,
            const_angle=45, radians=False,
            )
    params.update(opencl)

    si = LatticeSpinIce(basis0=b0, basis1=b1, **params)

    # change basis0
    b0_new = [2,1]
    si.set_basis(basis0=b0_new)

    expect = LatticeSpinIce(basis0=b0_new, basis1=b1, **params)
    assert_array_equal(si.pos, expect.pos)
    assert_array_equal(si.angle, expect.angle)

    # change basis1
    b1_new = [-.5,-1]
    si.set_basis(basis1=b1_new)

    expect = LatticeSpinIce(basis0=b0_new, basis1=b1_new, **params)
    assert_array_equal(si.pos, expect.pos)
    assert_array_equal(si.angle, expect.angle)

    # change both basis
    si = LatticeSpinIce(basis0=b0, basis1=b1, **params)
    si.set_basis(b0_new, b1_new)

    expect = LatticeSpinIce(basis0=b0_new, basis1=b1_new, **params)
    assert_array_equal(si.pos, expect.pos)
    assert_array_equal(si.angle, expect.angle)

def test_indexof():
    si = LatticeSpinIce(size=(4,3))

    inds = [si.indexof(tuple(l)) for l in si.labels]
    assert_array_equal(inds, si.indices())


def test_num_neighbors():
    size = (10,10)

    # nearest neighborhood (default)
    si = LatticeSpinIce(size=size)
    assert si.num_neighbors == 4

    # global neighborhood
    si = LatticeSpinIce(size=size, neighbor_distance=np.inf)
    assert si.num_neighbors == si.spin_count - 1


def test_neighbors():
    size=(3,3)
    si = LatticeSpinIce(size=size, basis0=[0.1,0.9], basis1=[-0.1,0.9])
    L = si.L

    # First "horizontal" row
    i = L[0,0]
    ns = set(si.neighbors(i))
    assert ns == {L[1, 0], L[0, 1]}

    i = L[0,1]
    ns = set(si.neighbors(i))
    assert ns == {L[0, 0], L[1, 0], L[2, 0], L[1, 1], L[0, 2]}

    # First vertical row
    i = L[1,0]
    ns = set(si.neighbors(i))
    assert ns == {L[0, 0], L[2, 0], L[0, 1], L[1, 1], L[0, 2]}

    i = L[1,1]
    ns = set(si.neighbors(i))
    assert ns == {L[1, 0], L[2, 0], L[0, 1], L[2, 1], L[0, 2], L[1, 2]}

    i = L[1,2]
    ns = set(si.neighbors(i))
    assert ns == {L[2, 0], L[1, 1], L[2, 1], L[0, 2], L[2, 2]}

    # Middle "horizontal" row
    i = L[2,0]
    ns = set(si.neighbors(i))
    assert ns == {L[1, 0], L[0, 1], L[1, 1], L[2, 1], L[0, 2], L[1, 2]}

    i = L[2,2]
    ns = set(si.neighbors(i))
    assert ns == {L[2,1], L[1,2]}


def test_spin_dipolar_field(opencl):
    # default lattice spacing = 1 such that NN islands have coupling strength 1.5
    size = (3,3)
    si = LatticeSpinIce(size=size, basis0=(1, 0), basis1=(0, 1), const_angle=0, **opencl)
    L = si.L

    a, b = 0.1767767, 0.53033009
    assert si.spin_dipolar_field(L[0,0], L[0,1]) == approx([-1, 0])
    assert si.spin_dipolar_field(L[0,0], L[1,0]) == approx([2, 0])
    assert si.spin_dipolar_field(L[0,0], L[1,1]) == approx([a,  b])
    assert si.spin_dipolar_field(L[0,0], L[2,0]) == approx([0.25, 0])

    assert si.spin_dipolar_field(L[1,0], L[0,0]) == approx([2, 0])
    assert si.spin_dipolar_field(L[1,0], L[0,1]) == approx([a, -b])

    assert si.spin_dipolar_field(L[1,1], L[0,1]) == approx([2, 0])
    assert si.spin_dipolar_field(L[1,1], L[0,2]) == approx([a, -b])
    assert si.spin_dipolar_field(L[1,1], L[2,1]) == approx([2, 0])
    assert si.spin_dipolar_field(L[1,1], L[2,2]) == approx([a, b])
    assert si.spin_dipolar_field(L[1,1], L[0,0]) == approx([a, b])
    assert si.spin_dipolar_field(L[1,1], L[2,0]) == approx([a, -b])

def test_dipolar_field(opencl):
    size = (3,3)
    si = LatticeSpinIce(size=size, alpha=1.0, neighbor_distance=1.5, basis0=(1, 0), basis1=(0, 1), const_angle=0, **opencl)
    L = si.L

    assert si.dipolar_field(L[1,1]) == approx([2.70710678, 0])
    si.flip(L[2,1])
    assert si.dipolar_field(L[1,1]) == approx([-1.29289322, 0])
    si.flip(L[0,2])
    assert si.dipolar_field(L[1,1]) == approx([-1.64644661,  1.06066017])

def test_external_field(opencl):
    size = (3,3)
    si = LatticeSpinIce(size=size, neighbor_distance=1, basis0=(1, 0), basis1=(0, 0.5), const_angle=30, **opencl)
    L = si.L
    a30 = np.deg2rad(30)
    cos30 = np.cos(a30)
    sin30 = np.sin(a30)
    cos60 = np.cos(np.deg2rad(60))
    sin60 = np.sin(np.deg2rad(60))

    assert si.external_field(L[0,0]) == approx([0, 0])
    assert si.external_field(L[1,0]) == approx([0, 0])

    # Horizontal
    si.set_h_ext((1,0))
    assert si.external_field(L[0,0]) == approx([cos30, -sin30])
    assert si.external_field(L[0,1]) == approx([cos30, -sin30])
    assert si.external_field(L[1,0]) == approx([cos30, -sin30])

    # Vertical
    si.set_h_ext((0,1))
    assert si.external_field(L[0,0]) == approx([sin30, cos30])
    assert si.external_field(L[0,1]) == approx([sin30, cos30])
    assert si.external_field(L[1,0]) == approx([sin30, cos30])

    # 30 degrees
    si.set_h_ext((cos30, sin30))
    assert si.external_field(L[0,0]) == approx([1, 0])
    assert si.external_field(L[0,1]) == approx([1, 0])
    assert si.external_field(L[1,0]) == approx([1, 0])

    # -30 degrees
    si.set_h_ext((np.cos(-a30), np.sin(-a30)))
    assert si.external_field(L[0,0]) == approx([sin30, -cos30])
    assert si.external_field(L[0,1]) == approx([sin30, -cos30])
    assert si.external_field(L[1,0]) == approx([sin30, -cos30])

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_flippable(opencl, switching):
    size = (4,3)
    si = LatticeSpinIce(size=size, neighbor_distance=1, basis0=(1, 0), basis1=(0, 0.5), const_angle=30, alpha=1.0, hc=1, switching=switching, **opencl)

    if switching == "budrikis":
        assert len(si.flippable()) == 8
    elif switching == "sw":
        assert len(si.flippable()) == 8

    si.flip(si.L[2,2])
    E = si.switching_energy()
    flippable = [i for i in si.indices() if E[i] > 0]
    assert len(flippable) > 0
    assert_array_equal(si.flippable(), flippable)

    si.polarize()
    si.set_h_ext([-20, 0])
    assert len(si.flippable()) == si.spin_count


def test_spin_grid(opencl):
    ls = 1/2
    si = LatticeSpinIce(size=(2,2), lattice_spacing=ls, **opencl)

    a30 = np.deg2rad(30)
    a90 = np.deg2rad(90)
    v30n = [-np.sin(-a30), -np.cos(-a30)]
    v30p = [np.sin(a30), np.cos(a30)]
    v90 = [1, 0]
    v0 = [0, 0]

    # grid_spacing == lattice_spacing
    grid = si.spin_grid()
    assert grid.shape == (2, 2, 2)
    expect = [[[1, 0], [1, 0]],
              [[1, 0], [1, 0]]]

    assert_allclose(grid, np.array(expect))



#!/usr/bin/env python3
import itertools
from os import path

from flatspin.data import Dataset, table_columns
from flatspin.cmdline import *

def main_inspect(dataset, args):
    for i, ds in dataset.items():
        print(f"[{i}] ", end='')
        print(" ".join(f"{k}={v}" for k,v in dict(ds.row()).items()), end='')
        print(":")

        tablefiles = ds.tablefiles()
        print(f"    tables ({len(tablefiles)}): ", end='')
        tabletypes = ["[t]" if "t" in table_columns(f) else "" for f in tablefiles]
        print(", ".join(f + t for (f, t) in zip(map(path.basename, tablefiles), tabletypes)))

        files = set(ds.files()) - set(tablefiles)
        if files:
            print(f"    files ({len(files)}): ", end='')
            files = sorted(files, key=lambda f: f[::-1])
            files = itertools.groupby(files, key=lambda f: path.splitext(f)[1][1:])
            print(", ".join(f"{ext} ({len(list(group))})" for ext, group in files if ext))

    return 0


# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = main_dataset_argparser("Inspect the contents of a dataset.")

def main():
    args = parser.parse_args()

    ds = main_dataset(args)

    return main_inspect(ds, args)

if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from .utils import approx, assert_allclose
from flatspin.grid import Grid

# Some test data: 10x6 grid
# Y/X
#    0  1  2  3  4  5  6  7  8  9
# 0 00 01 02 03 04 05 06 07 08 09
# 1 10 11 12 13 14 15 16 17 18 19
# 2 20 21 22 23 24 25 26 27 28 29
# 3 30 31 32 33 34 35 36 37 38 39
# 4 40 41 42 43 44 45 46 47 48 49
# 5 50 51 52 53 54 55 56 57 58 59
X, Y = np.meshgrid(np.arange(10.0), np.arange(6.0))
X = X.flatten()
Y = Y.flatten()
XY = np.column_stack([X, Y])
YX = np.flip(XY, axis=-1)

def test_cell_size():
    grid = Grid(XY, cell_size=(1,1))

    assert grid.cell_size == (1,1)
    assert grid.padding == (0.5, 0.5)
    assert grid.size == (6, 10)
    assert grid.extent == (-0.5, -0.5, 9.5, 5.5)

    assert_array_equal(grid._grid_index, YX)

    grid = Grid(XY, cell_size=1)
    assert grid.cell_size == (1,1)
    assert grid.padding == (0.5, 0.5)
    assert grid.size == (6, 10)
    assert grid.extent == (-0.5, -0.5, 9.5, 5.5)

    # auto cell size
    XY2 = XY.copy()
    XY2[:,0] = XY2[:,0] / 2
    XY2[:,1] = XY2[:,1] / 3
    grid = Grid(XY2)

    assert grid.cell_size[0] == approx(1/2)
    assert grid.cell_size[1] == approx(1/3)
    assert grid.padding[0] == approx(1/4)
    assert grid.padding[1] == approx(1/6)
    assert grid.size == (6, 10)
    assert grid.extent == approx((-1/4, -1/6, 9/2 + 1/4, 5/3 + 1/6))

    grid = Grid(XY, cell_size=(2,2), padding=(0.5, 0.5))

    assert grid.cell_size == (2,2)
    assert grid.padding == (0.5, 0.5)
    assert grid.size == (3, 5)
    assert grid.extent == (-0.5, -0.5, 9.5, 5.5)

    grid = Grid(XY, cell_size=(3, 3), padding=(0.5, 0.5))

    assert grid.cell_size == (3,3)
    assert grid.padding == (0.5, 0.5)
    assert grid.size == (2, 4)
    assert grid.extent == (-0.5, -0.5, 11.5, 5.5)

def test_fixed_grid():
    grid = Grid.fixed_grid(XY, (10, 6))

    assert grid.cell_size == (1, 1)
    assert grid.padding == (0.5, 0.5)
    assert grid.size == (6, 10)
    assert grid.extent == (-0.5, -0.5, 9.5, 5.5)

    grid = Grid.fixed_grid(XY, (5, 6))
    assert grid.cell_size == (2, 1)
    assert grid.padding == (0.5, 0.5)
    assert grid.size == (6, 5)
    assert grid.extent == (-0.5, -0.5, 9.5, 5.5)

    grid = Grid.fixed_grid(XY, (5, 3))
    assert grid.cell_size == (2, 2)
    assert grid.padding == (0.5, 0.5)
    assert grid.size == (3, 5)
    assert grid.extent == (-0.5, -0.5, 9.5, 5.5)

def test_center_grid():
    grid = Grid(XY, cell_size=(1,1))

    assert_array_equal(grid.center_grid(), XY.reshape(grid.size + (2,)))

    grid = Grid.fixed_grid(XY, (5, 6))
    xx, yy = np.meshgrid(0.5 + np.arange(0, 10.0, 2), np.arange(6.0))
    expect = np.stack([xx, yy], axis=-1)
    assert_array_equal(grid.center_grid(), expect)

    grid = Grid.fixed_grid(XY, (5, 3))
    xx, yy = np.meshgrid(0.5 + np.arange(0, 10.0, 2), 0.5 + np.arange(0, 6.0, 2))
    expect = np.stack([xx, yy], axis=-1)
    assert_array_equal(grid.center_grid(), expect)

def test_grid_index():
    grid = Grid(XY, cell_size=1)
    assert np.all([grid.grid_index(i) == YX[i] for i in range(len(YX))])

    gi, gj = grid.grid_index(np.arange(len(XY)))
    assert_array_equal(gi, Y.astype(int))
    assert_array_equal(gj, X.astype(int))

    grid = Grid(XY, cell_size=2, padding=0.5)
    # refer to table at the top of this file...
    assert grid.grid_index( 0) == grid.grid_index( 1) == (0,0)
    assert grid.grid_index(10) == grid.grid_index(11) == (0,0)
    assert grid.grid_index( 2) == grid.grid_index( 3) == (0,1)
    assert grid.grid_index(12) == grid.grid_index(13) == (0,1)
    assert grid.grid_index( 4) == grid.grid_index( 5) == (0,2)
    assert grid.grid_index(14) == grid.grid_index(15) == (0,2)
    assert grid.grid_index( 6) == grid.grid_index( 7) == (0,3)
    assert grid.grid_index(16) == grid.grid_index(17) == (0,3)
    assert grid.grid_index( 8) == grid.grid_index( 9) == (0,4)
    assert grid.grid_index(18) == grid.grid_index(19) == (0,4)

    assert grid.grid_index(20) == grid.grid_index(21) == (1,0)
    assert grid.grid_index(30) == grid.grid_index(31) == (1,0)
    assert grid.grid_index(22) == grid.grid_index(23) == (1,1)
    assert grid.grid_index(32) == grid.grid_index(33) == (1,1)
    assert grid.grid_index(24) == grid.grid_index(25) == (1,2)
    assert grid.grid_index(34) == grid.grid_index(35) == (1,2)
    assert grid.grid_index(26) == grid.grid_index(27) == (1,3)
    assert grid.grid_index(36) == grid.grid_index(37) == (1,3)
    assert grid.grid_index(28) == grid.grid_index(29) == (1,4)
    assert grid.grid_index(38) == grid.grid_index(39) == (1,4)

    assert grid.grid_index(40) == grid.grid_index(41) == (2,0)
    assert grid.grid_index(50) == grid.grid_index(51) == (2,0)
    assert grid.grid_index(42) == grid.grid_index(43) == (2,1)
    assert grid.grid_index(52) == grid.grid_index(53) == (2,1)
    assert grid.grid_index(44) == grid.grid_index(45) == (2,2)
    assert grid.grid_index(54) == grid.grid_index(55) == (2,2)
    assert grid.grid_index(46) == grid.grid_index(47) == (2,3)
    assert grid.grid_index(56) == grid.grid_index(57) == (2,3)
    assert grid.grid_index(48) == grid.grid_index(49) == (2,4)
    assert grid.grid_index(58) == grid.grid_index(59) == (2,4)

    grid = Grid(XY, cell_size=(5,3), padding=0.5)
    gi, gj = grid.grid_index([0,1,2,3,4,10,11,12,13,14,20,21,22,23,24])
    assert np.all(gi == 0)
    assert np.all(gj == 0)
    gi, gj = grid.grid_index([5,6,7,8,9,15,16,17,18,19,25,26,27,28,29])
    assert np.all(gi == 0)
    assert np.all(gj == 1)
    gi, gj = grid.grid_index([30,31,32,33,34,40,41,42,43,44,50,51,52,53,54])
    assert np.all(gi == 1)
    assert np.all(gj == 0)
    gi, gj = grid.grid_index([35,36,37,38,39,45,46,47,48,49,55,56,57,58,59])
    assert np.all(gi == 1)
    assert np.all(gj == 1)

def test_point_index():
    grid = Grid(XY, cell_size=1)
    grid_inds = np.transpose(grid.grid_index(np.arange(len(XY))))
    assert np.all([grid.point_index(gi) == i for i,gi in enumerate(grid_inds)])

    grid = Grid(XY, cell_size=2, padding=0.5)
    # refer to table at the top of this file...
    assert_array_equal(grid.point_index((0,0)), [0, 1, 10, 11])
    assert_array_equal(grid.point_index((0,1)), [2, 3, 12, 13])
    assert_array_equal(grid.point_index((0,2)), [4, 5, 14, 15])
    assert_array_equal(grid.point_index((0,3)), [6, 7, 16, 17])
    assert_array_equal(grid.point_index((0,4)), [8, 9, 18, 19])

    assert_array_equal(grid.point_index((1,0)), [20, 21, 30, 31])
    assert_array_equal(grid.point_index((1,1)), [22, 23, 32, 33])
    assert_array_equal(grid.point_index((1,2)), [24, 25, 34, 35])
    assert_array_equal(grid.point_index((1,3)), [26, 27, 36, 37])
    assert_array_equal(grid.point_index((1,4)), [28, 29, 38, 39])

    assert_array_equal(grid.point_index((2,0)), [40, 41, 50, 51])
    assert_array_equal(grid.point_index((2,1)), [42, 43, 52, 53])
    assert_array_equal(grid.point_index((2,2)), [44, 45, 54, 55])
    assert_array_equal(grid.point_index((2,3)), [46, 47, 56, 57])
    assert_array_equal(grid.point_index((2,4)), [48, 49, 58, 59])

    grid = Grid(XY, cell_size=(5,3), padding=0.5)
    assert_array_equal(grid.point_index((0,0)), [0,1,2,3,4,10,11,12,13,14,20,21,22,23,24])
    assert_array_equal(grid.point_index((0,1)), [5,6,7,8,9,15,16,17,18,19,25,26,27,28,29])
    assert_array_equal(grid.point_index((1,0)), [30,31,32,33,34,40,41,42,43,44,50,51,52,53,54])
    assert_array_equal(grid.point_index((1,1)), [35,36,37,38,39,45,46,47,48,49,55,56,57,58,59])

def test_map_values():
    grid = Grid(XY, cell_size=1)

    values = np.arange(len(XY))
    mapped = grid.map_values(values)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped.flatten(), values)

    # num values < grid
    values = np.arange(10)
    mapped = grid.map_values(values)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped[0], values)
    assert np.all(mapped[1:] == 0)

    mapped = grid.map_values(values, fill_value=-1)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped[0], values)
    assert np.all(mapped[1:] == -1)

    # mask empty cells
    values = np.arange(10)
    mapped = grid.map_values(values, mask_empty=True)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped[0], values)
    assert np.all(mapped[1:].data == 0)
    assert np.all(mapped.mask[0] == False)
    assert np.all(mapped.mask[1:] == True)

    # multiple values map to same grid cell
    grid = Grid(XY, cell_size=2, padding=0.5)
    values = np.arange(len(XY))
    mapped = grid.map_values(values)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped,
            [[11, 13, 15, 17, 19],
             [31, 33, 35, 37, 39],
             [51, 53, 55, 57, 59]])

    # vector values
    values = YX
    mapped = grid.map_values(values)
    assert mapped.shape == grid.size + (2,)
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped,
            [[[1,1], [1,3], [1,5], [1,7], [1,9]],
             [[3,1], [3,3], [3,5], [3,7], [3,9]],
             [[5,1], [5,3], [5,5], [5,7], [5,9]]])

def test_add_values():
    grid = Grid(XY, cell_size=1)

    values = np.arange(len(XY))
    mapped = grid.add_values(values)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped.flatten(), values)

    # num values < grid
    values = np.arange(10)
    mapped = grid.add_values(values)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped[0], values)
    assert np.all(mapped[1:] == 0)

    mapped = grid.add_values(values, fill_value=-1)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped[0], values-1)
    assert np.all(mapped[1:] == -1)

    # mask empty cells
    values = np.arange(10)
    mapped = grid.add_values(values, mask_empty=True)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    assert_array_equal(mapped[0], values)
    assert np.all(mapped[1:].data == 0)
    assert np.all(mapped.mask[0] == False)
    assert np.all(mapped.mask[1:] == True)

    # multiple values map to same grid cell
    grid = Grid(XY, cell_size=2, padding=0.5)
    values = np.arange(len(XY), dtype=float)
    mapped = grid.add_values(values)
    assert mapped.shape == grid.size
    assert mapped.dtype == values.dtype
    expect = np.array(
            [[0+1+10+11, 2+3+12+13, 4+5+14+15, 6+7+16+17, 8+9+18+19],
             [20+21+30+31, 22+23+32+33, 24+25+34+35, 26+27+36+37, 28+29+38+39],
             [40+41+50+51, 42+43+52+53, 44+45+54+55, 46+47+56+57, 48+49+58+59]],
            )
    assert_array_equal(mapped, expect)

    # mean
    mapped = grid.add_values(values, method='mean')
    assert_array_equal(mapped, expect/4)

    values = [9,3,5]
    mapped = grid.add_values(values, method='mean')
    assert mapped[0,0] == (9+3)/2
    assert mapped[0,1] == 5
    assert_array_equal(mapped.flatten()[2:], 0)

    # vector values
    values = YX
    mapped = grid.add_values(values)
    assert mapped.shape == grid.size + (2,)
    assert mapped.dtype == values.dtype

    assert tuple(mapped[0,0]) == (2,2)
    assert tuple(mapped[0,1]) == (2,10)
    assert tuple(mapped[0,2]) == (2,18)
    assert tuple(mapped[0,3]) == (2,26)
    assert tuple(mapped[0,4]) == (2,34)

    assert tuple(mapped[1,0]) == (10,2)
    assert tuple(mapped[1,1]) == (10,10)
    assert tuple(mapped[1,2]) == (10,18)
    assert tuple(mapped[1,3]) == (10,26)
    assert tuple(mapped[1,4]) == (10,34)

    assert tuple(mapped[2,0]) == (18,2)
    assert tuple(mapped[2,1]) == (18,10)
    assert tuple(mapped[2,2]) == (18,18)
    assert tuple(mapped[2,3]) == (18,26)
    assert tuple(mapped[2,4]) == (18,34)

    values = np.array([[1,2],[3,4],[5,6]])
    mapped = grid.add_values(values, method='mean')
    assert mapped.shape == grid.size + (2,)
    assert mapped.dtype == values.dtype
    assert tuple(mapped[0,0]) == ((1+3)/2, (2+4)/2)
    assert tuple(mapped[0,1]) == (5, 6)
    assert_array_equal(mapped.flatten()[4:], 0)

    # mask empty cells
    mapped = grid.add_values(values, mask_empty=True)
    assert tuple(mapped[0,0]) == (1+3, 2+4)
    assert tuple(mapped[0,1]) == (5, 6)
    assert_array_equal(mapped.flatten()[4:], 0)
    assert mapped.shape == mapped.mask.shape == grid.size + (2,)
    assert mapped.dtype == values.dtype
    assert np.all(mapped.mask[0,0] == False)
    assert np.all(mapped.mask[0,1] == False)
    assert np.all(mapped.mask.flatten()[4:] == True)



#!/usr/bin/env python3
import sys
import numpy as np
from numpy.linalg import norm
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from joblib import delayed

from flatspin.data import *
from flatspin.utils import *
from flatspin.cmdline import *
from flatspin.grid import Grid
from flatspin.plotting import montage_fig, vector_montage, save_animation, gridless_crop

def load_vector_data(dataset, args):
    quantity = args.quantity
    t = parse_time(args.t, numeric_dict(dataset.params))
    grid_size = args.grid
    crop_width = args.crop
    win_shape, win_step = args.window if args.window else (None, None)

    time, UV = read_vectors(dataset.tablefiles(), quantity, t)
    pos, angle = read_geometry(dataset.tablefile('geometry'))

    if UV.shape[1:] != pos.shape:
        # global field
        assert UV.shape[1:] == (1,2)
        UV = np.repeat(UV, len(pos), axis=1)

    if grid_size == (0, 0):
        # Grid disabled
        if args.crop:
            in_crop = gridless_crop(pos, args.crop)

            pos = pos[in_crop]
            UV = UV[:,in_crop]
        return time, pos, UV

    XY, UV = vector_grid(pos, UV, grid_size, crop_width, win_shape, win_step, normalize=False)

    return time, XY, UV




def main_load_data(dataset, args):
    # Do it in parallell
    progress_bar = ProgressBar(desc="Loading data", total=len(dataset))
    parallel = ParallelProgress(progress_bar, n_jobs=-1)

    data = parallel(
        delayed(load_vector_data)(ds, args)
        for ds in dataset)

    progress_bar.close()

    times, positions, vectors = tuple(zip(*data)) # unzip
    # TODO: assumes runs have the same number of timesteps
    times = np.array(times)

    # Normalize vectors to unit length
    nmax = np.max([norm(v.reshape((-1,2)), axis=-1).max() for v in vectors])
    if nmax != 0:
        vectors = [v / nmax for v in vectors]

    if args.compress:
        # Filter out states/timesteps where none of the runs change state
        print(f"Compressed {times.shape[1]} -> ", end="")

        # For each run, compare all vectors over time
        mask = [v[1:] == v[:-1] for v in vectors]

        # For each run, find states with no changes
        mask = [m.reshape((m.shape[0], -1)) for m in mask]
        mask = [~np.all(m, axis=-1) for m in mask]

        # Combine above information across runs:
        # keep time steps where any of the runs have changed state
        mask = np.any(mask, axis=0)

        # Always show first state
        mask = np.concatenate(([True], mask))

        # Filter out times/vectors where nothing has changed
        times = times[:, mask]
        vectors = [v[mask] for v in vectors]

        print(f"{times.shape[1]} frames")

    return times, positions, vectors

def main_vectors_anim_time(dataset, args):
    """ Animate over time, montage over dataset """
    param = args.param if args.param else dataset.index.columns[0]

    dataset = dataset.sort_values(param)

    times, positions, vectors = main_load_data(dataset, args)
    n_images = len(times)
    n_frames = len(times[0])

    label = args.label
    if label is None:
        if np.issubdtype(dataset.index[param].dtype, np.number):
            label = param + "={" + param + ":g}"
        else:
            label = param + "={" + param + "}"

    labels = None
    if label:
        labels = [label.format(**dict(row, i=i)) for i, row in dataset.index.iterrows()]

    title = args.title if args.title is not None else "t={t:g}"

    fig, axes = montage_fig(n_images, title or label)

    def init():
        pass

    def step(i):
        t = times[0][i]
        vec = [v[i] for v in vectors]

        if title:
            kw = {'name': dataset.name, 'basepath': dataset.basepath, 't': t}
            kw.update(dataset.params)
            kw.update(dataset.index.iloc[0])
            fig.suptitle(title.format(**kw))

        vector_montage(axes, positions, vec, labels, args.style, cmap=args.cmap)

    repeat = n_frames > 1
    ani = animation.FuncAnimation(fig, step, n_frames, init_func=init,
            interval=1000/args.fps, blit=False, save_count=sys.maxsize,
            repeat=repeat)

    return ani

def main_vectors_anim_dataset(dataset, args):
    """ Animate over dataset, montage over time """
    param = args.param if args.param else dataset.index.columns[0]

    dataset = dataset.sort_values(param)

    times, positions, vectors = main_load_data(dataset, args)

    n_images = len(vectors[0])
    n_frames = len(times)

    label = args.label if args.label is not None else "t={t:g}"
    title = args.title if args.title is not None else param + "={" + param + "}"

    fig, axes = montage_fig(n_images, title)

    def init():
        pass

    def step(i):
        time = times[i]
        XY = positions[i]
        UV = vectors[i]

        assert len(UV) == n_images

        if title:
            kw = {'name': dataset.name, 'basepath': dataset.basepath,
                  'index': dataset.index.index[i]}
            kw.update(dataset.params)
            kw.update(dataset.index.iloc[i])
            fig.suptitle(title.format(**kw))

        if label:
            labels = [label.format(t=t) for t in time]
        else:
            labels = None

        vector_montage(axes, [XY]*n_images, UV, labels, args.style, cmap=args.cmap)

    repeat = n_frames > 1
    ani = animation.FuncAnimation(fig, step, n_frames, init_func=init,
            interval=1000/args.fps, blit=False, save_count=sys.maxsize,
            repeat=repeat)

    return ani

def main_vectors(dataset, args):
    if args.dark:
        plt.style.use("dark_background")
    if args.grid == (0, 0):
        # Grid disabled
        if args.style == "image":
            raise ValueError("Image-based visualization not possible without a grid, try --style=arrow?")
        #if args.crop:
        #    raise ValueError("Can't --crop without a grid")
        if args.window:
            raise ValueError("Can't --window without a grid")

    if args.animate_dataset:
        ani = main_vectors_anim_dataset(dataset, args)
    else:
        ani = main_vectors_anim_time(dataset, args)

    if args.output:
        save_animation(ani, args.output, args.fps, args.dpi)
    else:
        plt.show()
    return 0


# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = main_dataset_grid_argparser("Plot vector data.")

parser.add_argument('-p', '--param', help='montage over given parameter')
parser.add_argument('-a', '--animate-dataset', action='store_true',
        help='animate over dataset (instead of time)')
parser.add_argument('-q', '--quantity', default='mag',
        help='quantity to view (default: %(default)s)')
parser.add_argument('--fps', type=float, default=30,
        help='frames per second (default: %(default)s)')
parser.add_argument('--dpi', type=int, default=100,
        help='dots per inch (default: %(default)s)')
parser.add_argument('--title', help='title format string', nargs='?', const="",)
parser.add_argument('--dark', action='store_true', help='dark mode')
parser.add_argument('--label', help='label format', nargs='?', const="",)
parser.add_argument('--style', choices=('image', 'arrow', 'rectangle', 'stadium'),
        default='image', help='vector style (default: %(default)s)')
parser.add_argument('--cmap', default='flatspin', help='set colormap')
parser.add_argument('--compress', action='store_true',
        help='filter time steps where there is no change in state')

def main():
    args = parser.parse_args()
    ds = main_dataset(args)
    return main_vectors(ds, args)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import numpy as np
from numpy.linalg import norm
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from skimage.util import view_as_windows
from joblib import delayed

from flatspin.data import *
from flatspin.utils import *
from flatspin.cmdline import *
from flatspin.grid import Grid
from flatspin.vertices import find_vertices, vertex_pos, vertex_mag

from . import vectors

def load_vertex_data(dataset, args):

    quantity = args.quantity
    t = parse_time(args.t, numeric_dict(dataset.params))
    # Not sure we need grid/crop/win for vertices, but can comment out the
    # lines below to enable it later
    #grid_size = args.grid
    #crop_width = args.crop
    #win_shape, win_step = args.window if args.window else (None, None)
    vertex_size = tuple(args.vertex_size)

    time, UV = read_vectors(dataset.tablefiles(), quantity, t)
    pos, angle = read_geometry(dataset.tablefile('geometry'))

    if UV.shape[1:] != pos.shape:
        # global field
        assert UV.shape[1:] == (1,2)
        UV = np.repeat(UV, len(pos), axis=1)

    # Map the vectors on the native grid
    grid = Grid(pos)

    vi, vj, vertices = find_vertices(grid, pos, angle, vertex_size)

    XY = vertex_pos(pos, vertices)
    UV = np.array([vertex_mag(UVi, vertices) for UVi in UV])

    # Apply vector grid on the vertices
    #XY, UV = vector_grid(XY, UV, grid_size, crop_width, win_step, normalize=False)

    return time, XY, UV


# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = main_dataset_argparser("Plot vertices.")
#parser = main_dataset_grid_argparser("Plot vertices")

parser.add_argument('-t', default='::spp',
        help='time range start:stop:step (default: %(default)s)')
parser.add_argument('--vertex-size', nargs=2, type=int, default=(3, 3),
        help='Size of vertex window')
parser.add_argument('--what', choices=('type', 'mag'), default='mag')
parser.add_argument('-p', '--param', help='montage over given parameter')
parser.add_argument('-a', '--animate-dataset', action='store_true',
        help='animate over dataset (instead of time)')
parser.add_argument('-q', '--quantity', default='mag',
        help='quantity to view (default: %(default)s)')
parser.add_argument('--fps', type=float, default=30,
        help='frames per second (default: %(default)s)')
parser.add_argument('--title', help='title format string')
parser.add_argument('--dark', action='store_true', help='dark mode')
parser.add_argument('--label', help='label format')
parser.add_argument('--style', choices=('arrow', 'rectangle', 'stadium'),
        default='arrow', help='vector style (default: %(default)s)')
parser.add_argument('--cmap', default='flatspin', help='set colormap')
parser.add_argument('--compress', action='store_true',
        help='filter time steps where there is no change in state')

def main():
    args = parser.parse_args()

    # Monkey patch vectors to load vertices :)
    vectors.load_vector_data = load_vertex_data
    # Grid, crop and window not supported (for now)
    args.grid = (0, 0)
    args.crop = None
    args.window = None

    ds = main_dataset(args)

    vectors.main_vectors(ds, args)

if __name__ == '__main__':
    main()



"""
Data management.
"""
import sys
import os
import copy
import numpy as np
from numpy.linalg import norm
import zipfile
import numpy.lib.format as npy_format
import pandas as pd
from glob import glob
from os import path
import fnmatch
import re
from collections import OrderedDict
from datetime import timedelta
from skimage.util import view_as_windows
# pd.set_option('display.max_colwidth', -1)
pd.set_option('display.max_columns', None)

from .utils import eval_dict
from .grid import Grid

def col_group(x):
    if x[-1] in 'xyz':
        return x[:-1]
    if ',' in x:
        return x.split(',')[0]
    return x

def match_column(pattern, columns):
    """
    match column pattern from available columns
    m -> (mx, my, mz)
    m.region*x
    """
    cols = []
    if pattern in columns:
        cols.append(pattern)
    elif pattern + "x" in columns:
        cols.append(pattern + "x")
        if pattern + "y" in columns:
            cols.append(pattern + "y")
        if pattern + "z" in columns:
            cols.append(pattern + "z")
    else:
        matches = fnmatch.filter(columns, pattern)
        if matches:
            cols.extend(matches)
    return cols

def match_columns(patterns, columns):
    """ match column patterns """
    cols = []
    for p in patterns:
        cols.extend(match_column(p, columns))
    return cols

def csv_columns(filename):
    """ Parse column names at the top of a CSV file """
    with open(filename) as fp:
        first_line = fp.readline()

    if not first_line.startswith('#'):
        # no header found
        return []

    first_line = first_line.strip('# \n')
    if '(' in first_line:
        # some CSV headers contain extra info such as units per column
        # e.g. "t (s) B_extx (T) ..."
        # remove them from the column names
        first_line = re.sub(r' \([^()]*\)', '', first_line)

    columns = first_line.split()

    return columns

def save_csv(df, filename):
    index = df.index.name is not None
    header = not isinstance(df.columns, pd.RangeIndex)

    fp = open(filename, 'wt')
    if header:
        fp.write('# ')

    # For object columns, df.to_csv uses str(obj) to get a string
    # representation of each value. In case the column contains a numpy array,
    # make sure the entire array is returned, without newlines.
    np.set_string_function(repr, repr=False) # use repr() as the str() function
    with np.printoptions(linewidth=np.inf, threshold=np.inf):
        df.to_csv(fp, sep=' ', index=index, header=header)
    np.set_string_function(None, repr=False) # restore default str() function

def read_csv(filename, index_col=None):
    try:
        names = csv_columns(filename)
        if names:
            df = pd.read_csv(filename, sep=r'\s+', index_col=index_col,
                    names=names, skiprows=1, engine='c')
        else:
            df = pd.read_csv(filename, sep=r'\s+', index_col=index_col,
                    header=None, engine='c')

    except (pd.errors.EmptyDataError, pd.errors.ParserError) as e:
        # empty csv
        df = pd.DataFrame()

    return df

def save_hdf(df, filename):
    filename, key = path.split(filename)
    df.to_hdf(filename, key, mode='a', complevel=9, complib='zlib')

def read_hdf(filename):
    filename, key = archive_key(filename)
    return pd.read_hdf(filename, key)

def hdf_columns(filename):
    filename, key = archive_key(filename)
    store = pd.HDFStore(filename, mode='r')

    if key is None:
        key = store.keys()[0]

    # assuming fixed format
    columns = list(store.get_node('{}/axis0'.format(key)).read().astype(str))
    index_name = store.get_node('{}/axis1'.format(key))._v_attrs.name
    if index_name is not None:
        columns.append(index_name)

    store.close()

    return columns

def read_npy(filename):
    arr = np.load(filename, allow_pickle=True)
    return pd.DataFrame.from_records(arr)

def save_npy(df, filename):
    index = df.index.name is not None
    arr = df.to_records(index=index)
    np.save(filename, arr)

def npy_columns(filename):
    with open(filename, 'rb') as fp:
        version = npy_format.read_magic(fp)
        npy_format._check_version(version)
        shape, fortran_order, dtype = npy_format._read_array_header(fp, version)
        return dtype.names

def read_npz(filename):
    filename, key = archive_key(filename)
    with np.load(filename, allow_pickle=True) as f:
        return pd.DataFrame.from_records(f[key])

def save_npz(df, filename):
    filename, key = archive_key(filename)
    index = df.index.name is not None
    arr = df.to_records(index=index)

    """ np.savez() doesn't support appending to an archive,
    so we need to do it ourselves """
    kw = dict(compression=zipfile.ZIP_DEFLATED, allowZip64=True)
    with zipfile.ZipFile(filename, mode="a", **kw) as zipf:
        fname = key + '.npy'
        # TODO: check for duplicate keys
        # always force zip64, gh-10776
        with zipf.open(fname, 'w', force_zip64=True) as fid:
            npy_format.write_array(fid, arr, allow_pickle=True)

def npz_columns(filename):
    filename, key = archive_key(filename)
    with np.load(filename) as npz:
        fp = npz.zip.open(key + ".npy")
        version = npy_format.read_magic(fp)
        npy_format._check_version(version)
        try:
            # numpy 1.23.5 introduced a safety check for large headers, which
            # needs to be disabled for larger tables by increasing max_header_size
            shape, fortran_order, dtype = npy_format._read_array_header(fp, version,
                    max_header_size=2**64)
        except TypeError:
            # numpy versions older than 1.23.5 do not understand max_header_size
            shape, fortran_order, dtype = npy_format._read_array_header(fp, version)
        return dtype.names

# supported table formats
table_formats = {
    'csv': (read_csv, save_csv, csv_columns),
    'hdf': (read_hdf, save_hdf, hdf_columns),
    'npy': (read_npy, save_npy, npy_columns),
    'npz': (read_npz, save_npz, npz_columns),
}

# map from extension -> format
table_extensions = {
    'csv': 'csv',
    'txt': 'csv',
    'hdf': 'hdf',
    'h5': 'hdf',
    'npy': 'npy',
    'npz': 'npz',
}

archive_formats = ["npz", "hdf"]

table_patterns = ['*.' + ext for ext in table_extensions]

def archive_key(filename):
    base, ext = path.splitext(filename)
    if ext[1:] in archive_formats:
        return filename, None

    head, tail = path.split(filename)
    if not head:
        return tail, None

    return head, tail

def get_format(filename):
    # dir/file.ext
    base, ext = path.splitext(filename)
    if ext:
        return ext[1:]

    # archive.fmt/key
    filename, key = archive_key(filename)
    base, ext = path.splitext(filename)
    if ext:
        return ext[1:]

    # unknown
    return None

def is_archive_format(fmt):
    return fmt in archive_formats

def is_archive(filename):
    fmt = get_format(filename)
    return is_archive_format(fmt)

def is_tablefile(filename):
    return get_format(filename) in table_extensions

def list_dir(filename):
    return os.listdir(filename)

def list_npz(filename):
    with np.load(filename) as npz:
        return npz.files

def list_hdf(filename):
    with pd.HDFStore(filename, mode='r') as store:
        return [k.lstrip('/') for k in store.keys()]

def listfiles(filename):
    files = None

    if not path.exists(filename):
        raise FileNotFoundError(filename)

    if path.isdir(filename):
        files = list_dir(filename)

    elif is_archive(filename):
        fmt = get_format(filename)
        if fmt == 'npz':
            files = list_npz(filename)
        elif fmt == 'hdf':
            files = list_hdf(filename)

    if files:
        files = [path.join(filename, f) for f in sorted(files)]
        return files

    raise ValueError(f"Don't know how to open {filename}, sorry")

def to_table(data):
    if isinstance(data, dict):
        with np.printoptions(linewidth=np.inf, threshold=sys.maxsize):
            data = [[k, repr(v)] for k,v in data.items()]
        df = pd.DataFrame(data)
        return df
        # return pd.DataFrame(data.values(), index=data.keys())

    if not isinstance(data, pd.DataFrame):
        return pd.DataFrame(data)

    return data

def match_any(name, patterns):
    return any([fnmatch.fnmatch(name, pat) for pat in patterns])

def read_table(filename, index_col=None, **kwargs):
    fmt = get_format(filename)
    fmt = table_extensions[fmt]
    reader = table_formats[fmt][0]
    df = reader(filename, **kwargs)
    if index_col is not None and df.index.name != index_col:
        df.set_index(index_col, inplace=True)
    return df

def save_table(data, filename):
    df = to_table(data)
    fmt = get_format(filename)
    fmt = table_extensions[fmt]
    writer = table_formats[fmt][1]
    return writer(df, filename)

def table_columns(filename):
    fmt = get_format(filename)
    fmt = table_extensions[fmt]
    func = table_formats[fmt][2]
    return func(filename)

def read_tables(filenames, index_col=None):
    dfs = [read_table(f, index_col) for f in filenames]

    # merge them
    table = dfs[0]
    for df in dfs[1:]:
        table = table.join(df)

    return table

def read_geometry(filename='geometry.csv'):
    df = read_table(filename)
    pos = np.array(df[['posx', 'posy']])
    angle = np.array(df['angle'])
    return pos, angle

def read_vectors(filenames, quantity='mag', t=None):
    files = []
    columns = []
    for f in filenames:
        cols = table_columns(f)
        # Look for time column
        if 't' not in cols:
            continue

        # Look for quantity
        match = match_column(quantity + '*', cols)
        if not match:
            continue

        # Look for xy data
        match = list(filter(lambda c: c[-1] in 'xy', match))
        if not match:
            continue

        files.append(f)
        columns.extend(match)

    assert files, f"No vector data found for quantity: {quantity}"

    df = read_tables(files, index_col='t')
    df = df[columns]

    if t is not None:
        if np.isscalar(t):
            t = [t]
        df = df.iloc[t]

    assert len(df.columns) % 2 == 0, "Number of columns not divisible by 2, not vector data?"

    UV = np.array(df)
    UV = UV.reshape((UV.shape[0], -1, 2))

    time = np.array(df.index)

    return time, UV

def crop(X, crop, axis=-1):
    """ Crop array X along one or more axes
    crop is a tuple (before, after) for each axis to crop
    axis specifies which axis to start cropping from
    """
    if not isinstance(X, np.ndarray):
        X = np.array(X)

    if np.isscalar(crop):
        crop = (crop, crop)

    crop = np.array(crop)
    if crop.ndim < 2:
        crop = [crop] * len(X.shape)

    if axis < 0:
        axis = X.ndim - len(crop) + axis + 1

    indices = [slice(None) for i in range(X.ndim)]
    for i, (before, after) in enumerate(crop):
        start = before
        stop = -after if after > 0 else None
        indices[axis + i] = slice(start, stop)

    return X[tuple(indices)]

def rolling_window(X, win_shape, step=1, method='sum'):
    """ Apply a function over a rolling window of X """
    assert method in ('sum', 'mean')

    mask = X.mask if np.ma.is_masked(X) else None
    X = np.copy(X)
    dtype = X.dtype
    if mask is not None:
        # view_as_windows doesn't work with masked arrays so set masked values
        # to nan and ignore the nans when computing the mean
        X = X.astype(float)
        X[mask] = np.nan

    if np.isscalar(win_shape):
        win_shape = (win_shape,) * X.ndim

    if np.isscalar(step):
        step = (step,) * X.ndim

    assert len(win_shape) == len(step)

    W = view_as_windows(X, win_shape, step)
    axis = tuple(range(X.ndim, W.ndim))

    if mask is not None:
        if method == 'mean':
            W = np.nanmean(W, axis=axis)
        else:
            W = np.nansum(W, axis=axis)

        # Mask any remaining nan values
        W = np.ma.masked_invalid(W)
        return W.astype(dtype)

    if method == 'mean':
        return np.mean(W, axis=axis)

    return np.sum(W, axis=axis)


def filter_df(df, **kwargs):
    """ Filter dataframe by key=value or range key=slice(start, stop, step) or list key=[key1, key2, key3] """
    for k,v in kwargs.items():
        if callable(v):
            df = df[df[k].apply(v)]

        elif isinstance(v, slice):
            # range (start, stop)
            start = v.start - 1e-8 if v.start is not None else v.start
            stop = v.stop + 1e-8 if v.stop is not None else v.stop
            step = v.step
            df = df.set_index(k).loc[start:stop:step].reset_index()
        else:
            if not isinstance(v, tuple) and not isinstance(v, list):
                v = [v]

            matches = []
            for vi in v:
                # equals
                if isinstance(vi, str) or np.issubdtype(df[k].dtype, np.integer):
                    matches.append(df[k] == vi)
                else:
                    # Determine tolerance based on size of filter value
                    atol = 10**(np.floor(np.log10(np.abs(vi))) - 4)
                    matches.append(np.isclose(df[k], vi, atol=atol))

            df = df[np.logical_or.reduce(matches)]

    return df

class Dataset(object):
    def __init__(self, index=None, params={}, info={}, basepath=None):
        self.index = index
        self.params = params
        self.info = info
        self.basepath = basepath

    @property
    def name(self):
        if self.basepath:
            return path.basename(path.abspath(self.basepath))
        return ''

    def __getitem__(self, i):
        index = self.index.loc[i]
        if isinstance(index, pd.Series):
            # single row returns a Series, but we want a DataFrame
            index = self.index.loc[[i]]
        ds = copy.copy(self)
        ds.index = index
        return ds

    def __repr__(self):
        return 'Dataset({!r}): {} items'.format(self.basepath, len(self))

    def __str__(self):
        s = "Dataset: {}\n\n".format(self.name)
        s += "params:\n"
        for k,v in sorted(self.params.items(), key=lambda i: i[0].lower()):
            s += " {}={}\n".format(k,v)
        s += "\n"
        s += "info:\n"
        for k,v in sorted(self.info.items(), key=lambda i: i[0].lower()):
            s += " {}: {}\n".format(k,v)
        s += "\n"
        s += "index:\n"
        s += str(self.index)
        s += "\n"
        return s

    def keys(self):
        return list(self.index.index)

    def items(self):
        for i in self.index.index:
            yield i, self.__getitem__(i)

    def iterrows(self):
        for i, row in self.index.iterrows():
            yield ((i,) + tuple(row), self.__getitem__(i))

    def __iter__(self):
        for _, ds in self.items():
            yield ds

    def __len__(self):
        return len(self.index)

    def __eq__(self, other):
        return self.basepath == other.basepath and \
                self.params == other.params and \
                (self.index == other.index).all(axis=None)

    def subset(self, i):
        index = self.index.loc[i]
        if isinstance(index, pd.Series):
            # single row returns a Series, but we want a DataFrame
            index = self.index.loc[[i]]
        ds = copy.copy(self)
        ds.index = index
        return ds

    def filter(self, **kwargs):
        ds = copy.copy(self)
        ds.index = filter_df(self.index, **kwargs)

        return ds

    def drop_duplicates(self, **kwargs):
        ds = copy.copy(self)
        ds.index = self.index.drop_duplicates(**kwargs)

        return ds

    def groupby(self, key):
        for k, index in self.index.groupby(key):
            ds = copy.copy(self)
            ds.index = index
            yield k, ds

    def sort_values(self, column):
        ds = copy.copy(self)
        ds.index = ds.index.sort_values(column)
        return ds

    def row(self, row=0):
        return self.index.iloc[row]

    def id(self, row=0):
        return self.row(row).name

    @staticmethod
    def read(basepath):
        index = read_csv(path.join(basepath, 'index.csv'))

        params = read_csv(path.join(basepath, 'params.csv'))
        params = dict(np.array(params))
        params = eval_dict(params)

        info = read_csv(path.join(basepath, 'info.csv'))
        info = dict(np.array(info))
        info = eval_dict(info)

        return Dataset(index, params, info, basepath)

    def save(self, basepath=None):
        if basepath and not self.basepath:
            self.basepath = basepath

        if not basepath:
            basepath = self.basepath

        assert basepath

        save_table(self.index, path.join(basepath, 'index.csv'))
        save_table(self.params, path.join(basepath, 'params.csv'))
        save_table(self.info, path.join(basepath, 'info.csv'))

    def file(self, filename):
        files = []
        for outdir in self.index['outdir']:
            files.append(path.join(self.basepath, outdir, filename))
        if len(files) == 1:
            return files[0]
        return files

    def files(self, patterns=None, squash=True):
        if isinstance(patterns, str):
            patterns = [patterns]

        files = [listfiles(path.join(self.basepath, outdir))
                 for outdir in self.index['outdir']]

        if patterns:
            files = [list(filter(lambda f: match_any(path.basename(f), patterns), fs))
                    for fs in files]

        if len(files) == 1 and squash:
            return files[0]

        return files

    def tablefile(self, tablename, squash=True):
        files = self.files(tablename + '*', squash=False)
        # next() to get the first matching tablefile
        files = [next(filter(is_tablefile, fs)) for fs in files]

        if len(files) == 1 and squash:
            return files[0]

        return files

    def tablefiles(self, patterns=None, squash=True):
        files = self.files(patterns, squash=False)
        files = [list(filter(is_tablefile, fs)) for fs in files]

        if len(files) == 1 and squash:
            return files[0]

        return files

#
# Data conversion
#
def digitize(X, threshold=0):
    return np.where(X > threshold, 1, 0)

def bit_array(x, n_bits):
    """ Convert number x to array of bits """
    return np.array([(x & (1 << bit)) >> bit for bit in range(n_bits)])

def array_bit(a):
    """ Convert bit array a to number """
    return sum([(int(b) << bit) for (bit, b) in enumerate(a)])

def vector_grid(pos, vectors, grid_size=None, crop_width=None,
        win_shape=None, win_step=None, normalize=True, return_grid=False):
    """ Process vectors on a grid """

    # Map the vectors on a grid
    if grid_size:
        # fixed grid width x height
        grid = Grid.fixed_grid(pos, grid_size)
    else:
        # native grid
        grid = Grid(pos)

    XY = grid.center_grid()

    # Determine whether we have a time axis
    assert len(vectors.shape) in (3, 2)
    assert vectors.shape[-1] == 2
    time_axis = 1 if len(vectors.shape) == 3 else 0

    if time_axis:
        # vectors.shape = (time, spin_count, 2)
        # UV.shape = (time, height, width, 2)
        UV = np.array([grid.add_values(UVi) for UVi in vectors])
    else:
        # vectors.shape = (spin_count, 2)
        # UV.shape = (height, width, 2)
        UV = grid.add_values(vectors)

    if crop_width:
        XY = crop(XY, crop_width, axis=0)
        UV = crop(UV, crop_width, axis=time_axis)

    if win_shape:
        if not win_step:
            win_step = win_shape

        # XY.shape = (height, width, 2)
        win_shape += (1,)
        win_step += (1,)
        XY = rolling_window(XY, win_shape, win_step, method='mean')

        if time_axis:
            # UV.shape = (time, height, width, 2)
            win_shape = (1,) + win_shape
            win_step = (1,) + win_step

        UV = rolling_window(UV, win_shape, win_step, method='sum')

    if normalize:
        # Normalize vectors to unit length
        nmax = norm(UV.reshape((-1,2)), axis=-1).max()
        if nmax != 0:
            UV = UV / nmax

    if return_grid:
        return XY, UV, grid

    return XY, UV

#
# Data loading utilities
#
def load_output(dataset, quantity, t=None, grid_size=None, crop_width=None,
        win_shape=None, win_step=None, flatten=True):
    """ Load output vectors from dataset, with optional post-processing """

    assert len(dataset) == 1

    time, UV = read_vectors(dataset.tablefiles(), quantity, t)

    if grid_size or crop_width or win_shape:
        XY, angle = read_geometry(dataset.tablefile('geometry'))
        XY, UV = vector_grid(XY, UV, grid_size, crop_width, win_shape, win_step)

    if flatten:
        UV = UV.reshape((UV.shape[0], -1))

    return UV

def load_input(dataset, t=None, input_column='input'):
    """ Load input data from dataset """

    assert len(dataset) == 1

    df = read_table(dataset.tablefile('input'), index_col='t')
    df = df[input_column]

    if t is not None:
        if np.isscalar(t):
            t = [t]
        df = df.iloc[t]

    return np.array(df)



from setuptools_scm import get_version
try:
    # If flatspin source is a git repo, try to determine version at runtime
    __version__ = get_version()
except LookupError:
    # Fall back to version detected at package installation time
    from ._version import version as __version__

from .model import *
from .data import Dataset
from .grid import Grid



import pytest
from flatspin.cmdline import parse_filter

def test_parse_filter():
    parsed_filter = parse_filter("1")
    assert parsed_filter == 1
    assert isinstance(parsed_filter, int)

    parsed_filter =  parse_filter("1.0")
    assert parsed_filter == 1.0
    assert isinstance(parsed_filter, float)

    assert parse_filter("(1,2,3)") == (1,2,3)
    assert parse_filter("[1,2,3]") == [1,2,3]

    assert parse_filter("1::") == slice(1, None, None)
    assert parse_filter("1:") == slice(1, None, None)
    assert parse_filter("1:2:3") == slice(1, 2, 3)
    assert parse_filter("1:2:") == slice(1, 2, None)

    assert parse_filter("float.is_integer") == float.is_integer
    assert callable(parse_filter("lambda x: x>5"))





import numpy as np
import pandas as pd
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal

from flatspin.encoder import *

def test_scale():
    input = np.array([0, 0.5, 1])
    out = scale(input, 0.5, 1)
    assert_array_equal(out, [0.5, 0.75, 1.0])

def test_broadcast_waveform():
    # 1d input
    input = [0, 1, 2]
    waveform = [0, 1, 0]
    out = broadcast_waveform(input, waveform)
    assert_array_equal(out, [0, 0, 0, 0, 1, 0, 0, 2, 0])

    # single 1d input
    input = [2]
    waveform = [0, 1, 0]
    out = broadcast_waveform(input, waveform)
    assert_array_equal(out, [0, 2, 0])

    # 2d input
    input = [[[0], [1]],
             [[2], [3]]]
    out = broadcast_waveform(input, waveform)
    expect = [[0, 0],
              [0, 1],
              [0, 0],
              [0, 0],
              [2, 3],
              [0, 0]]
    assert_array_equal(out, expect)

    # 1d input signal
    input = [[0, 1, 2],
             [3, 4, 5]]
    out = broadcast_waveform(input, waveform)
    assert_array_equal(out, [0, 1, 0, 0, 4, 0])

    # 2d input signal
    input = [[[0, 1, 2], [3, 4, 5]],
             [[6, 7, 8], [9, 10, 11]]]
    out = broadcast_waveform(input, waveform)
    expect = [[0, 0],
              [1, 4],
              [0, 0],
              [0, 0],
              [7, 10],
              [0, 0]]
    assert_array_equal(out, expect)

def test_sin():
    s = np.sin(np.linspace(0, 2 * np.pi, 100, endpoint=False))

    # 3 inputs, each with single feature
    input = np.array([1, 2, 3])
    input.shape += (1,)
    out = sin(input, 100)
    assert out.shape == (300,)
    expect = np.concatenate([1*s, 2*s, 3*s])
    assert_array_equal(out, expect)

    # 2 inputs, each with 3 features
    input = np.array([[1, 2, 3],
                      [4, 5, 6]])
    input.shape += (1,)
    out = sin(input, 100)
    assert out.shape == (200, 3)
    expect = [np.transpose([1*s, 2*s, 3*s]),
              np.transpose([4*s, 5*s, 6*s])]
    expect = np.concatenate(expect)
    assert_array_equal(out, expect)

    # 2 inputs, each with 2x2 features
    input = np.array([[[1, 2],
                       [3, 4]],
                      [[5, 6],
                       [7, 8]]])
    input.shape += (1,)
    out = sin(input, 100)
    assert out.shape == (200, 2, 2)
    expect = [[[1*s, 2*s],
               [3*s, 4*s]],
              [[5*s, 6*s],
               [7*s, 8*s]]]
    expect = np.rollaxis(np.array(expect), -1, 1)
    expect = np.concatenate(expect)
    assert_array_equal(out, expect)

def test_rotate():
    c = np.cos(np.linspace(0, 2 * np.pi, 100, endpoint=False))
    s = np.sin(np.linspace(0, 2 * np.pi, 100, endpoint=False))

    # 1d input
    input = [1, 2, 3]
    out = rotate(input, 100)
    assert out.shape == (300, 2)
    expect1 = np.concatenate([1*c, 2*c, 3*c])
    expect2 = np.concatenate([1*s, 2*s, 3*s])
    expect = np.column_stack([expect1, expect2])
    assert_array_equal(out, expect)

    # 2 inputs, each with 3 features
    input = np.array([[1, 2, 3],
                      [4, 5, 6]])
    input.shape += (1,)
    out = rotate(input, 100)
    assert out.shape == (200, 3, 2)
    expect0 = [np.concatenate([1*c, 4*c]), np.concatenate([1*s, 4*s])]
    assert_array_equal(out[:,0], np.column_stack(expect0))
    expect1 = [np.concatenate([2*c, 5*c]), np.concatenate([2*s, 5*s])]
    assert_array_equal(out[:,1], np.column_stack(expect1))
    expect2 = [np.concatenate([3*c, 6*c]), np.concatenate([3*s, 6*s])]
    assert_array_equal(out[:,2], np.column_stack(expect2))

    # 2 inputs, each with 2x2 features
    input = np.array([[[1, 2],
                       [3, 4]],
                      [[5, 6],
                       [7, 8]]])
    input.shape += (1,)
    out = rotate(input, 100)
    assert out.shape == (200, 2, 2, 2)
    expect00 = [np.concatenate([1*c, 5*c]), np.concatenate([1*s, 5*s])]
    assert_array_equal(out[:,0,0], np.column_stack(expect00))
    expect11 = [np.concatenate([4*c, 8*c]), np.concatenate([4*s, 8*s])]
    assert_array_equal(out[:,1,1], np.column_stack(expect11))

def test_angle():
    # 3 inputs, each with single feature
    input = np.array([0, 0.25, 0.5, 0.75, 1])
    out = angle(input)
    assert out.shape == (5, 2)
    expect = [[1, 0],
              [0, 1],
              [-1, 0],
              [0, -1],
              [1, 0]]
    assert_allclose(out, expect)

    # 2 inputs, each with 3 features
    input = np.array([[0, 0.25, 0.5],
                      [0.75, 1.0, 0.5]])
    out = angle(input)
    assert out.shape == (2, 3, 2)
    expect = [[[1, 0], [0, 1], [-1, 0]],
              [[0, -1], [1, 0], [-1, 0]]]
    assert_allclose(out, expect)

def test_fixed_vector():
    # 3 inputs, each with single feature
    input = np.array([1, 2, 3])
    c = np.cos(np.deg2rad(30))
    s = np.sin(np.deg2rad(30))
    out = fixed_vector(input, 30)
    assert out.shape == (3, 2)
    expect = [[1*c, 1*s],
              [2*c, 2*s],
              [3*c, 3*s]]
    assert_array_equal(out, expect)

    # 2 inputs, each with 3 features
    input = np.array([[1, 2, 3],
                      [4, 5, 6]])
    out = fixed_vector(input, 30)
    assert out.shape == (2, 3, 2)
    expect = [[[1*c, 1*s], [2*c, 2*s], [3*c, 3*s]],
              [[4*c, 4*s], [5*c, 5*s], [6*c, 6*s]]]
    assert_array_equal(out, expect)

def test_rotate_vector():
    def v(deg):
        rad = np.deg2rad(deg)
        return np.column_stack([np.cos(rad), np.sin(rad)])

    #
    # 1d array of vectors
    #
    degs = np.linspace(0, 360, 5*5)
    input = np.array(v(degs))

    # 0 deg rotation -> no change
    out = rotate_vector(input, 0)
    assert_array_equal(out, input)

    # 90 deg rotation
    out = rotate_vector(input, 90)
    expect = v(degs + 90)
    assert_allclose(out, expect)

    # -90 deg rotation
    out = rotate_vector(input, -90)
    expect = v(degs - 90)
    assert_allclose(out, expect)

    # 60 deg rotation
    out = rotate_vector(input, 60)
    expect = v(degs + 60)
    assert_allclose(out, expect)

    # 360 + 60 deg rotation
    out = rotate_vector(input, 360+60)
    expect = v(degs + 60)
    assert_allclose(out, expect)

    #
    # 2d array of vectors
    #
    input = input.reshape((5,5,2))

    # 0 deg rotation -> no change
    out = rotate_vector(input, 0)
    assert_array_equal(out, input)

    # 90 deg rotation
    out = rotate_vector(input, 90)
    expect = v(degs + 90).reshape(input.shape)
    assert_allclose(out, expect)

    # -90 deg rotation
    out = rotate_vector(input, -90)
    expect = v(degs - 90).reshape(input.shape)
    assert_allclose(out, expect)

    # 60 deg rotation
    out = rotate_vector(input, 60)
    expect = v(degs + 60).reshape(input.shape)
    assert_allclose(out, expect)

    # 360 + 60 deg rotation
    out = rotate_vector(input, 360+60)
    expect = v(degs + 60).reshape(input.shape)
    assert_allclose(out, expect)


def test_ensure2d():
    #
    # 1d scalar inputs -> vectors
    #
    input = np.array([0,1,2,3])

    # default: fill y component with 1
    out = ensure2d(input)
    expect = [[0,1], [1,1], [2,1], [3,1]]
    assert_array_equal(out, expect)

    # fill y component with 5
    out = ensure2d(input, fill=5)
    expect = [[0,5], [1,5], [2,5], [3,5]]
    assert_array_equal(out, expect)

    # vector inputs -> unchanged
    input = np.array([[0,1],[2,3]])
    out = ensure2d(input)
    expect = [[0,1], [2,3]]
    assert_array_equal(out, expect)

def test_grid():
    # 2 inputs, each with single feature
    input = np.array([1, 2])
    out = grid(input, (3, 2))
    assert out.shape == (2, 2, 3)
    expect = [[[1,1,1],
               [1,1,1]],
              [[2,2,2],
               [2,2,2]]]
    assert_array_equal(out, expect)

    # 2 inputs, each with 2 features
    input = np.array([[1, 2], [3, 4]])
    out = grid(input, (3, 2))
    assert out.shape == (2, 2, 3, 2)
    expect = [[[[1,2],[1,2],[1,2]],
               [[1,2],[1,2],[1,2]]],
              [[[3,4],[3,4],[3,4]],
               [[3,4],[3,4],[3,4]]]]
    assert_array_equal(out, expect)

    # 1D grid
    input = np.array([1, 2])
    out = grid(input, (3, 1))
    assert out.shape == (2, 1, 3)
    expect = [[[1,1,1]],
              [[2,2,2]]]
    assert_array_equal(out, expect)

    input = np.array([1, 2])
    out = grid(input, (1, 3))
    assert out.shape == (2, 3, 1)
    expect = [[[1],
               [1],
               [1]],
              [[2],
               [2],
               [2]]]
    assert_array_equal(out, expect)

def test_onehot():
    # single feature
    input = np.array([0, 1])
    out = onehot(input, nbits=2)
    assert_array_equal(out, [1, 0, 0, 1])

    input = np.array([0, 1])
    out = onehot(input, nbits=3)
    assert_array_equal(out, [1, 0, 0, 0, 1, 0])

    input = np.array([0, 1, 2])
    out = onehot(input, nbits=3)
    assert_array_equal(out, [1, 0, 0, 0, 1, 0, 0, 0, 1])

    # 2 inputs, each with 2 features
    input = np.array([[0, 1], [2, 0]])
    out = onehot(input, nbits=3)

             # 0, 1
    expect = [[1, 0],
              [0, 1],
              [0, 0],

             # 2, 0
              [0, 1],
              [0, 0],
              [1, 0]]
    assert_array_equal(out, expect)

def test_ppm():
    # single feature
    input = np.array([0, 0.5, 1])
    out = ppm(input, timesteps=3)
    expect = [[1,0,0],
              [0,1,0],
              [0,0,1]]
    assert_array_equal(out, expect)

    out = ppm(input, timesteps=6, pos=0.5)
    expect = [[1,0,0,0,0,0],
              [0,1,0,0,0,0],
              [0,0,1,0,0,0]]
    assert_array_equal(out, expect)

    out = ppm(input, timesteps=6, pos0=0.5)
    expect = [[0,0,0,1,0,0],
              [0,0,0,0,1,0],
              [0,0,0,0,0,1]]
    assert_array_equal(out, expect)

    # 2 inputs, each with 2 features
    input = np.array([[0, 0.5], [0.5, 1.0]])
    out = ppm(input, timesteps=3)
    expect = [[[1,0,0],[0,1,0]],
              [[0,1,0],[0,0,1]]]
    assert_array_equal(out, expect)

def test_constant_encoder():
    input = np.array([[0,.5,1]]).T
    encoder = Constant()

    out = encoder(input)
    expect = np.array([[0,0],[.5,0],[1,0]])
    assert_array_equal(out, expect)

    encoder.set_params(H=3)
    out = encoder(input)
    expect = np.array([[0,0],[1.5,0],[3,0]])
    assert_array_equal(out, expect)

    encoder.set_params(H0=1)
    out = encoder(input)
    expect = np.array([[1,0],[2,0],[3,0]])
    assert_array_equal(out, expect)

    encoder.set_params(phi=90)
    out = encoder(input)
    expect = np.array([[0,1],[0,2],[0,3]])
    assert_allclose(out, expect)

    encoder.set_params(timesteps=2)
    out = encoder(input)
    expect = np.array([[0,1],[0,1],[0,2],[0,2],[0,3],[0,3]])
    assert_allclose(out, expect)

def test_sin_encoder():
    input = np.array([[0,.5,1]]).T
    encoder = Sine()

    out = encoder(input)
    s = np.sin(np.linspace(0, 2*np.pi, 100, endpoint=False))
    expect = np.concatenate([0*s, .5*s, 1*s])
    assert_array_equal(out[:,0], expect)
    assert_array_equal(out[:,1], np.zeros(len(expect)))

    encoder.set_params(H=10)
    out = encoder(input)
    assert_array_equal(out[:,0], 10 * expect)
    assert_array_equal(out[:,1], np.zeros(len(expect)))

    encoder.set_params(H0=5)
    out = encoder(input)
    expect = np.concatenate([5*s, 7.5*s, 10*s])
    assert_array_equal(out[:,0], expect)
    assert_array_equal(out[:,1], np.zeros(len(expect)))

    encoder.set_params(phi=90)
    out = encoder(input)
    assert_allclose(out[:,0], np.zeros(len(expect)))
    assert_array_equal(out[:,1], expect)

def test_triangle_encoder():
    input = np.array([[0,.5,1]]).T
    encoder = Triangle(timesteps=8)

    out = encoder(input)
    s = np.array([0, 0.5, 1, 0.5, 0, -0.5, -1, -0.5])
    expect = np.concatenate([0*s, .5*s, 1*s])
    assert_array_equal(out[:,0], expect)
    assert_array_equal(out[:,1], np.zeros(len(expect)))

    encoder.set_params(H=10)
    out = encoder(input)
    assert_array_equal(out[:,0], 10 * expect)
    assert_array_equal(out[:,1], np.zeros(len(expect)))

    encoder.set_params(H0=5)
    out = encoder(input)
    expect = np.concatenate([5*s, 7.5*s, 10*s])
    assert_array_equal(out[:,0], expect)
    assert_array_equal(out[:,1], np.zeros(len(expect)))

    encoder.set_params(phi=90)
    out = encoder(input)
    assert_allclose(out[:,0], np.zeros(len(expect)))
    assert_array_equal(out[:,1], expect)

def test_constant_grid_encoder():
    input = np.array([[0,.5,1]]).T
    encoder = ConstantGrid(grid_size=(3,3))

    out = encoder(input)
    s = np.ones((3,3))
    expect = np.concatenate([[0*s], [.5*s], [1*s]])
    assert_array_equal(out[...,0], expect)
    assert_array_equal(out[...,1], 0*expect)

    encoder.set_params(H=10)
    out = encoder(input)
    assert_array_equal(out[...,0], 10*expect)
    assert_array_equal(out[...,1], 0*expect)

    encoder.set_params(H0=5)
    out = encoder(input)
    expect = np.concatenate([[5*s], [7.5*s], [10*s]])
    assert_array_equal(out[...,0], expect)
    assert_array_equal(out[...,1], 0*expect)

    encoder.set_params(phi=90)
    out = encoder(input)
    assert_allclose(out[...,0], 0*expect)
    assert_array_equal(out[...,1], expect)

    encoder.set_params(timesteps=2)
    out = encoder(input)
    rexpect = np.repeat(expect, 2, axis=0)
    assert_allclose(out[...,0], 0*rexpect)
    assert_array_equal(out[...,1], rexpect)

    # custom grid weights
    grid = np.array([[0,1],[.5,0]])
    encoder.set_params(grid=grid, H0=0, H=1, phi=0, timesteps=1)

    out = encoder(input)
    expect = np.concatenate([[0*grid], [.5*grid], [1*grid]])
    assert_array_equal(out[...,0], expect)
    assert_array_equal(out[...,1], 0*expect)

    # scaling happens after the grid
    encoder.set_params(H0=5, H=10)
    out = encoder(input)
    expect = np.array([[[5,5],[5,5]],
                       [[5,7.5],[6.25,5]],
                       [[5,10],[7.5,5]]])
    assert_array_equal(out[...,0], expect)
    assert_array_equal(out[...,1], 0*expect)

    encoder.set_params(timesteps=2)
    out = encoder(input)
    rexpect = np.repeat(expect, 2, axis=0)
    assert_array_equal(out[...,0], rexpect)
    assert_array_equal(out[...,1], 0*rexpect)



#!/usr/bin/env python3
import sys
import os
import argparse
import shlex
import numpy as np
import pandas as pd
from collections import OrderedDict
from copy import copy

from flatspin import __version__
from flatspin.cmdline import StoreKeyValue, eval_params, main_version
from flatspin.sweep import sweep
from flatspin.data import Dataset, table_formats, is_archive_format, read_table, eval_dict
from flatspin.runner import run, run_local, run_dist
from flatspin import runner
from flatspin.utils import get_default_params, import_class
from .run import main_list_params, main_list_encoders, main_help_encoder, visualize_run

# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = argparse.ArgumentParser(description='Run flatspin sweep.')
parser.add_argument('-r', '--run', choices=['local', 'dist', 'none'], default='local',
                    help='run locally or distributed on a cluster')
parser.add_argument('-p', '--param', action=StoreKeyValue, default={},
                    help='set model parameter key=value')
parser.add_argument('-s', '--sweep', action=StoreKeyValue, metavar='key=SPEC',
                    help='set sweep parameter key=SPEC')
parser.add_argument('-n', '--repeat', type=int, default=1, metavar='N',
                    help='repeat each experiment N times (default: %(default)s)')
parser.add_argument('-ns', '--repeat-spec', action=StoreKeyValue, metavar='key=SPEC',
                    help='repeat each according to key=SPEC')
parser.add_argument('--seed', type=int, default=0, help='random seed')
parser.add_argument('-m', '--model', default='SquareSpinIceClosed',
        help='name of model (default: %(default)s) ')
parser.add_argument('-e', '--encoder', default='Sine',
        help='input encoder (type of external field, default: %(default)s)')
parser.add_argument('-o', '--basepath', help='output directory for results')
parser.add_argument('-f', '--format', choices=table_formats.keys(), default='npz',
        help='format of output files (default: %(default)s)')
parser.add_argument('--params-file',
        help='load parameters from file (params.csv), loaded before -p')
parser.add_argument('--sweep-file', help='load sweep (index.csv) from file')
parser.add_argument('--list-params', action='store_true')
parser.add_argument('--list-encoders', action='store_true')
parser.add_argument('--help-encoder', action='store_true')
parser.add_argument('-V', '--version', help='display flatspin version', action='store_true')
parser.add_argument('--show-plot', help='plot the state as the simulation runs', action='store_true')

def main():
    args = parser.parse_args()

    if args.version:
        return main_version(args)
    if args.list_params:
        return main_list_params(args)

    if args.list_encoders:
        if args.encoder == parser.get_default('encoder'):
            args.encoder = 'flatspin.encoder'
        return main_list_encoders(args)

    if args.help_encoder:
        return main_help_encoder(args)

    if args.basepath is None or (args.sweep is None and args.sweep_file is None):
        parser.error('The following arguments are required: -o/--basepath, -s/--sweep/--sweep-file')

    if args.sweep and args.sweep_file:
        parser.error('Cannot have both -s/--sweep and --sweep-file')

    model_class = import_class(args.model, 'flatspin.model')
    model_name = model_class.__name__
    encoder_class = import_class(args.encoder, 'flatspin.encoder')

    # -e Sine is really just an alias for -p encoder=Sine
    params = get_default_params(run)
    params['encoder'] = f'{encoder_class.__module__}.{encoder_class.__name__}'
    params.update(get_default_params(model_class))
    params.update(get_default_params(encoder_class))

    if args.params_file:
        prms = read_table(args.params_file)
        prms = dict(np.array(prms))
        prms = eval_dict(prms)
        params.update(prms)

    params.update(eval_params(args.param))

    info = {
        'model': f'{model_class.__module__}.{model_class.__name__}',
        'model_name': model_name,
        'data_format': args.format,
        'command': ' '.join(map(shlex.quote, sys.argv)),
        'version': __version__,
    }


    ext = args.format if is_archive_format(args.format) else "out"
    if args.repeat > 1 or args.repeat_spec:
        outdir_tpl = model_name + ".{:06d}.{:06d}." + ext
    else:
        outdir_tpl = model_name + ".{:06d}." + ext

    basepath = args.basepath

    if args.sweep_file:
        # Load index from file
        index = read_table(args.sweep_file)
        if not 'outdir' in index:
            # Generate outdir column
            outdirs = list(map(outdir_tpl.format, range(len(index))))
            index['outdir'] = outdirs

    else:
        # Generate index
        index = []

        for i, j, sweep_params in sweep(args.sweep, args.repeat, args.repeat_spec, params, args.seed):

            outdir = outdir_tpl.format(i, j)
            outpath = os.path.join(basepath, outdir)

            row = OrderedDict(sweep_params)
            row.update({'outdir': outdir})
            index.append(row)

        index = pd.DataFrame(index)

    if os.path.exists(basepath):
        # Refuse to overwrite an existing dataset
        raise FileExistsError(basepath)
    os.makedirs(basepath)

    # Save dataset
    dataset = Dataset(index, params, info, basepath)
    dataset.save()

    # Run!
    print("Starting sweep with {} runs".format(len(dataset)))
    if args.run == 'local':
        if args.show_plot:
            runner.run_sample_callback = visualize_run
        run_local(dataset)

    elif args.run == 'dist':
        run_dist(dataset)

if __name__ == '__main__':
    main()
